from copy import deepcopy
from decimal import Decimal
from typing import Optional
from langchain.tools import tool
from langchain.agents import AgentExecutor

from .data_converter import DataConverter
from h_adminsim.registry import STATUS_CODES
from h_adminsim.utils import log
from h_adminsim.utils.fhir_utils import *
from h_adminsim.utils.common_utils import (
    group_consecutive_segments,
    convert_segment_to_time,
    convert_time_to_segment,
    init_result_dict,
    compare_iso_time,
    get_iso_time,
)



class SchedulingRule:
    def __init__(self, 
                 metadata: dict, 
                 department_data: dict, 
                 environment, 
                 fhir_intergration: bool = False):
        self.environment = environment
        self._current_time = self.environment.current_time
        self._utc_offset = self.environment._utc_offset
        self._metadata = metadata
        self._department_data = department_data
        self._START_HOUR = self._metadata.get('time').get('start_hour')
        self._END_HOUR = self._metadata.get('time').get('end_hour')
        self._TIME_UNIT = self._metadata.get('time').get('interval_hour')
        self.current_time = environment.current_time
        self.fhir_integration = fhir_intergration


    def physician_filter(self, filtered_doctor_information: dict, preferred_doctor: str) -> list[str]:
        """
        Filter schedules by preferred doctor.

        Args:
            filtered_doctor_information (dict): Filtered doctor information after department filtering.
            preferred_doctor (str): The identifier of the preferred doctor.

        Returns:
            list[str]: A set of candidate schedules that match the preferred doctor.
        """
        candidate_schedules = set()
        schedule_info = filtered_doctor_information['doctor'][preferred_doctor]
        schedule_candidates = schedule_info['schedule']
        min_time_slot_n = int(Decimal(str(schedule_info['outpatient_duration'])) / Decimal(str(self._TIME_UNIT)))
        dates = sorted(list(schedule_candidates.keys()))
        for date in dates:
            schedule = schedule_candidates[date]
            fixed_schedule_segments = sum([convert_time_to_segment(self._START_HOUR, 
                                                                    self._END_HOUR, 
                                                                    self._TIME_UNIT, 
                                                                    fs) for fs in schedule], [])
            all_time_segments = convert_time_to_segment(self._START_HOUR, self._END_HOUR, self._TIME_UNIT)
            free_time = [s for s in range(len(all_time_segments)) if s not in fixed_schedule_segments]
            
            if len(free_time):
                valid_time_segments = [seg for seg in group_consecutive_segments(free_time) if len(seg) >= min_time_slot_n]
                for valid_time in valid_time_segments:
                    for i in range(len(valid_time) - min_time_slot_n + 1):
                        time_slot = valid_time[i:i+min_time_slot_n]
                        free_max_st, _ = convert_segment_to_time(self._START_HOUR, self._END_HOUR, self._TIME_UNIT, [time_slot[0]])
                        free_max_st_iso = get_iso_time(free_max_st, date, utc_offset=self._utc_offset)
                        if compare_iso_time(free_max_st_iso, self._current_time):
                            candidate_schedules.add(f"{preferred_doctor};;;{free_max_st_iso}")

        return list(candidate_schedules)
    

    def date_filter(self, filtered_doctor_information: dict, valid_date: str) -> list[str]:
        """
        Filter schedules by valid date.

        Args:
            filtered_doctor_information (dict): Filtered doctor information after department filtering.
            valid_date (str): The valid date from which to consider schedules.

        Returns:
            list[str]: A set of candidate schedules that are on or after the valid date.
        """
        candidate_schedules = set()
        schedule_infos = filtered_doctor_information['doctor']

        for doctor, schedule_info in schedule_infos.items():
            min_time_slot_n = int(Decimal(str(schedule_info['outpatient_duration'])) / Decimal(str(self._TIME_UNIT)))
            dates = sorted(list(schedule_info['schedule'].keys()))

            for date in dates:
                if not compare_iso_time(valid_date, date):
                    schedule = schedule_info['schedule'][date]
                    fixed_schedule_segments = sum([convert_time_to_segment(self._START_HOUR, 
                                                                            self._END_HOUR, 
                                                                            self._TIME_UNIT, 
                                                                            fs) for fs in schedule], [])
                    all_time_segments = convert_time_to_segment(self._START_HOUR, self._END_HOUR, self._TIME_UNIT)
                    free_time = [s for s in range(len(all_time_segments)) if s not in fixed_schedule_segments]
                    
                    if len(free_time):
                        valid_time_segments = [seg for seg in group_consecutive_segments(free_time) if len(seg) >= min_time_slot_n]
                        for valid_time in valid_time_segments:
                            for i in range(len(valid_time) - min_time_slot_n + 1):
                                time_slot = valid_time[i:i+min_time_slot_n]
                                free_max_st, _ = convert_segment_to_time(self._START_HOUR, self._END_HOUR, self._TIME_UNIT, [time_slot[0]])
                                free_max_st_iso = get_iso_time(free_max_st, date, utc_offset=self._utc_offset)
                                if compare_iso_time(free_max_st_iso, self._current_time):
                                    candidate_schedules.add(f"{doctor};;;{free_max_st_iso}")

        return list(candidate_schedules)
    

    def get_all(self, filtered_doctor_information: dict) -> list[str]:
        """
        Get all candidate schedules without any filtering.

        Args:
            filtered_doctor_information (dict): Filtered doctor information after department filtering.

        Returns:
            list[str]: A set of all candidate schedules without any filtering.
        """
        candidate_schedules = set()
        schedule_infos = filtered_doctor_information['doctor']

        for doctor, schedule_info in schedule_infos.items():
            min_time_slot_n = int(Decimal(str(schedule_info['outpatient_duration'])) / Decimal(str(self._TIME_UNIT)))
            dates = sorted(list(schedule_info['schedule'].keys()))

            for date in dates:
                schedule = schedule_info['schedule'][date]
                fixed_schedule_segments = sum([convert_time_to_segment(self._START_HOUR, 
                                                                        self._END_HOUR, 
                                                                        self._TIME_UNIT, 
                                                                        fs) for fs in schedule], [])
                all_time_segments = convert_time_to_segment(self._START_HOUR, self._END_HOUR, self._TIME_UNIT)
                free_time = [s for s in range(len(all_time_segments)) if s not in fixed_schedule_segments]
                
                if len(free_time):
                    valid_time_segments = [seg for seg in group_consecutive_segments(free_time) if len(seg) >= min_time_slot_n]
                    for valid_time in valid_time_segments:
                        for i in range(len(valid_time) - min_time_slot_n + 1):
                            time_slot = valid_time[i:i+min_time_slot_n]
                            free_max_st, _ = convert_segment_to_time(self._START_HOUR, self._END_HOUR, self._TIME_UNIT, [time_slot[0]])
                            free_max_st_iso = get_iso_time(free_max_st, date, utc_offset=self._utc_offset)
                            if compare_iso_time(free_max_st_iso, self._current_time):
                                candidate_schedules.add(f"{doctor};;;{free_max_st_iso}")

        return list(candidate_schedules)
    

    def find_idx(self, patient_schedule_list: list[dict], patient_name: str, doctor_name: str, date: str) -> int:
        """
        Identify the index of the appointment corresponding to the patient's request
        (e.g., cancellation or modification) from the patient's schedule list.

        Args:
            patient_schedule_list (list[dict]): A list of the patient's scheduled appointments.
                                                Each item contains appointment details such as doctor name, date, and time.
            patient_name (str): Name of the patient making the request.
            doctor_name (str): Name of the doctor associated with the target appointment.
            date (str): Date of the target appointment (YYYY-MM-DD).

        Returns:
            int: The index of the appointment that matches the patient's request.
        """
        for idx, patient_schedule in enumerate(patient_schedule_list):
            if patient_schedule['status'] == 'scheduled' and \
                patient_schedule['patient'].lower() == patient_name.lower() and \
                    patient_schedule['attending_physician'].lower() == doctor_name.lower() \
                        and patient_schedule['date'] == date:
                return idx
        return -1


    def find_earliest_time(self, schedules: list[str], delimiter: str = ';;;') -> dict:
        """
        Find the earliest schedule from the list of schedules.

        Args:
            schedules (list[str]): A list of schedules in the format "doctor;;;iso_time".
            delimiter (str, optional): The delimiter used to split doctor and iso_time. Defaults to ';;;'.

        Returns:
            dict: A dictionary containing the earliest doctor(s) and their corresponding schedule(s).
        """
        earliest_doctor, earliest_time = list(), list()

        for schedule in schedules:
            doctor, iso_time = schedule.split(delimiter)

            # skip when the slot is earlier than the current time
            if not compare_iso_time(iso_time, self.current_time):
                continue

            if not len(earliest_doctor):
                earliest_doctor.append(doctor) 
                earliest_time.append(iso_time)
                continue
            
            # Append if the iso_time is same with the alrealdy appended one
            if earliest_time[0] == iso_time:
                earliest_doctor.append(doctor)
                earliest_time.append(iso_time)
            
            elif compare_iso_time(earliest_time[0], iso_time):
                earliest_doctor = [doctor]
                earliest_time = [iso_time]
        
        return {'doctor': earliest_doctor, 'schedule': earliest_time}
    

    def cancel_schedule(self,
                        idx: int,
                        doctor_info: dict,
                        cancelled_schedule: dict) -> dict:
        """
        Cancel the schedule both in doctor_info and FHIR system.

        Args:
            idx (int): The index of the appointment to be cancelled.
            doctor_info (dict): The doctor information containing schedules.
            cancelled_schedule (dict): The schedule details to be cancelled.

        Returns:
            dict: Updated doctor information after cancellation.
        """
        doctor, date, time = cancelled_schedule['attending_physician'], cancelled_schedule['date'], cancelled_schedule['schedule']
        schedule_list = doctor_info[doctor]['schedule'][date]

        # Remove from doctor_information
        schedule_list.remove(time)  # In-place logic

        # Remove from FHIR
        if self.fhir_integration:
            fhir_appointment = DataConverter.get_fhir_appointment(data={'metadata': deepcopy(self._metadata),
                                                                        'department': deepcopy(self._department_data),
                                                                        'information': deepcopy(cancelled_schedule)})
            self.environment.delete_fhir({'Appointment': fhir_appointment})
        
        # Remove from environment patient_schedules
        self.environment.schedule_cancel_event(idx, True)

        return doctor_info


def create_tools(rule: SchedulingRule, 
                 doctor_info: dict,
                 patient_schedule_list: Optional[list[dict]] = None,
                 gt_idx: Optional[int] = None,
                 only_schedule_tool: bool = False) -> list[tool]:
    @tool
    def physician_filter_tool(preferred_doctor: str) -> dict:
        """
        Return the earliest available schedule for a preferred doctor.
        
        Args:
            preferred_doctor: Name of the preferred doctor

        Returns:
            dict: The earliest physician-filtered time slot and its information.
        """
        log(f'[TOOL CALL] physician_filter_tool | preferred_doctor={preferred_doctor}', color=True)
        prefix = 'Dr.'
        if prefix not in preferred_doctor:
            preferred_doctor = f'{prefix} {preferred_doctor}'
        schedules = rule.physician_filter(doctor_info, preferred_doctor)
        schedule = rule.find_earliest_time(schedules)
        return schedule

    @tool
    def date_filter_tool(valid_date: str) -> dict:
        """
        Return the earliest available schedule after a specific date.
        
        Args:
            valid_date: Date in YYYY-MM-DD format.
        
        Returns:
            dict: The earliest date-filtered time slot and its information.
        """
        log(f'[TOOL CALL] date_filter_tool | valid_date={valid_date}', color=True)
        schedules = rule.date_filter(doctor_info, valid_date)
        schedule = rule.find_earliest_time(schedules)
        return schedule

    @tool
    def get_all_time_tool() -> dict:
        """
        Return the earliest available schedule among the all available time slots.

        Returns:
            dict: The earliest time slot and its information.
        """
        log(f'[TOOL CALL] get_all_time_tool', color=True)
        schedules = rule.get_all(doctor_info)
        schedule = rule.find_earliest_time(schedules)
        return schedule

    @tool
    def cancel_tool(patient_name: str, doctor_name: str, date: str) -> dict:
        """
        Identify the index of the appointment to be cancelled from the patient's schedule list.

        Args:
            patient_name (str): Name of the patient requesting the cancellation.
            doctor_name (str): Name of the doctor for the appointment to be cancelled.
            date (str): Date of the appointment to be cancelled (YYYY-MM-DD).

        Returns:
            dict: A dictionary containing the cancelled_schedule, result_dict, and updated_doctor_info.
        """
        log(f'[TOOL CALL] cancel_tool | patient_name={patient_name}, doctor_name={doctor_name}, date={date}', color=True)
        result_dict, updated_doctor_info, cancelled_schedule = init_result_dict(), None, None
        prefix = 'Dr.'
        if prefix not in doctor_name:
            doctor_name = f'{prefix} {doctor_name}'
        index = rule.find_idx(patient_schedule_list, patient_name, doctor_name, date)

        # Update result_dict
        if gt_idx is None:
            result_dict['gt'].append({'cancel': None})
            result_dict['pred'].append({'cancel': index})
            result_dict['status'].append(None)
            result_dict['status_code'].append(None)
        else:
            status = True if index == gt_idx else False
            status_code = STATUS_CODES['correct'] if index == gt_idx else STATUS_CODES['cancel']['identify']
            result_dict['gt'].append({'cancel': gt_idx})
            result_dict['pred'].append({'cancel': index})
            result_dict['status'].append(status)
            result_dict['status_code'].append(status_code)

        # Update the schedule only when the cancellation is correct or there is no gt_idx
        if gt_idx is None or status:
            cancelled_schedule = patient_schedule_list[index]
            updated_doctor_info = rule.cancel_schedule(index, doctor_info, cancelled_schedule)
                
        return {'cancelled_schedule': cancelled_schedule, 'result_dict': result_dict, 'updated_doctor_info': updated_doctor_info}
    

    @tool
    def reschedule_tool(patient_name: str, doctor_name: str, date: str) -> dict:
        """
        Identify the index of the appointment to be rescheduled from the patient's schedule list.

        Args:
            patient_name (str): Name of the patient requesting the rescheduling.
            doctor_name (str): Name of the doctor for the appointment to be rescheduled.
            date (str): Date of the original appointment to be rescheduled (YYYY-MM-DD).

        Returns:
            dict: A dictionary containing the original_schedule and result_dict.
        """
        log(f'[TOOL CALL] reschedule_tool | patient_name={patient_name}, doctor_name={doctor_name}, date={date}', color=True)
        result_dict, original_schedule = init_result_dict(), None
        prefix = 'Dr.'
        if prefix not in doctor_name:
            doctor_name = f'{prefix} {doctor_name}'
        index = rule.find_idx(patient_schedule_list, patient_name, doctor_name, date)
        
        # Update result_dict
        if gt_idx is None:
            result_dict['gt'].append({'reschedule': None})
            result_dict['pred'].append({'reschedule': index})
            result_dict['status'].append(None)
            result_dict['status_code'].append(None)
        else:
            status = True if index == gt_idx else False
            status_code = STATUS_CODES['correct'] if index == gt_idx else STATUS_CODES['reschedule']['identify']
            result_dict['gt'].append({'reschedule': gt_idx})
            result_dict['pred'].append({'reschedule': index})
            result_dict['status'].append(status)
            result_dict['status_code'].append(status_code)

        if gt_idx is None or status:
            original_schedule = patient_schedule_list[index]

        return {'original_schedule': original_schedule, 'result_dict': result_dict}
    
    if only_schedule_tool:
        return [physician_filter_tool, date_filter_tool, get_all_time_tool]
    return [physician_filter_tool, date_filter_tool, get_all_time_tool, cancel_tool, reschedule_tool]



def scheduling_tool_calling(client: AgentExecutor, 
                            user_prompt: str,
                            history: list = []) -> dict:
    """
    Make an appointment using tool-calling agent.

    Args:
        client (AgentExecutor): The agent executor to handle tool calls.
        user_prompt (str): User prompt used for tool calling.
        history (list, optional): A list of LangChain HumanMessage and AIMessage objects. Defaults to [].
    
    Returns:
        dict: A dictionary containing the scheduled doctor and their corresponding schedule.
    """
    inputs = {
        "input": user_prompt,
        "chat_history": history,
    }
    response = client.invoke(inputs)
    steps = response.get("intermediate_steps") or []

    if len(steps) > 0:
        tool_output = steps[0][1]
        return {"type": "tool", "result": tool_output, "raw": response}

    # No tool call happened
    text = response.get("output") or ""
    return {"type": "text", "result": text, "raw": response}



import re

from h_adminsim.utils.common_utils import (
    iso_to_hour,
    iso_to_date,
    sort_schedule,
    convert_time_list_to_merged_time,
)



def sanitize_id(s: str) -> str:
    """
    Sanitize a string to conform to the pattern: ^[A-Za-z0-9\-\.]{1,64}$

    Args:
        s (str): The input string to sanitize.

    Returns:
        str: A sanitized string containing only allowed characters,
             and no longer than 64 characters.
    """
    cleaned = re.sub(r'[^A-Za-z0-9\-\.]', '', s)
    return cleaned[:64]



def get_individual_id(hospital: str, department_code: str, individual_name: str) -> str:
    """
    Make an individual ID.

    Args:
        hospital (str): A hospital name.
        department_code (str): A department code.
        individual_name (str): An individual name.
    
    Returns:
        str: A sanitized individual ID.
    """
    return sanitize_id(f'{hospital}-{department_code.lower()}-{individual_name}')



def get_practitionerrole_id(individual_id: str) -> str:
    """
    Make a practitioner role ID for an individual.

    Args:
        individual_id (str): An individual ID.

    Returns:
        str: A practitioner role ID.
    """
    return f'{individual_id}-role'



def get_schedule_id(individual_id: str) -> str:
    """
    Make a schedule ID for an individual.

    Args:
        individual_id (str): An individual ID.

    Returns:
        str: A schedule ID.
    """
    return f'{individual_id}-schedule'



def get_slot_id(individual_id: str, date: str, time_segment_index: int) -> str:
    """
    Make a slot ID for an individual.

    Args:
        individual_id (str): An individual ID.
        date (str): A date in ISO format (YYYY-MM-DD).
        time_segment_index (int): An index of start time segment.

    Returns:
        str: A slot ID.
    """
    return f"{individual_id}-{date.replace('-', '')}-slot{time_segment_index}"



def get_appointment_id(individual_id: str, date: str, start_time_segment_index: int, end_time_segment_index: int) -> str:
    """
    Make an appointment ID for an individual.

    Args:
        individual_id (str): An individual ID.
        date (str): A date in ISO format (YYYY-MM-DD).
        start_time_segment_index (int): An index of start time segment.
        end_time_segment_index (int): An index of end time segment.

    Returns:
        str: An appointment ID.
    """
    return f"{individual_id}-{date.replace('-', '')}-appn{start_time_segment_index}-{end_time_segment_index}"



def get_all_doctor_info(practitioners: list[dict],
                        practitioner_roles: list[dict],
                        schedules: list[dict],
                        slots: list[dict],
                        appointments: list[dict],
                        **kwargs) -> dict:
    """
    Make a current state of doctoral information based on the FHIR server.

    Args:
        practitioners (list[dict]): Practitioner resources currently used in the hospital environment of the simulation.
        practitioner_roles (list[dict]): PractitionerRole resources currently used in the hospital environment of the simulation.
        schedules (list[dict]): Schedule resources currently used in the hospital environment of the simulation.
        slots (list[dict]): Slot resources currently used in the hospital environment of the simulation._
        appointments (list[dict]): Appointment resources currently used in the hospital environment of the simulation.

    Returns:
        dict: Current state of doctoral information. 
    """
    # Prepare several pre-required data
    doctor_information = dict()
    practitioner_ref_to_role = dict()
    practitioner_ref_to_schedules = dict()
    practitioner_ref_to_name = {
        f"Practitioner/{practitioner['resource']['id']}": \
            f"{practitioner['resource']['name'][0]['prefix'][0]} {practitioner['resource']['name'][0]['given'][0]} {practitioner['resource']['name'][0]['family']}" \
                for practitioner in practitioners
    }
    for practitioner_role in practitioner_roles:
        attributes = {attr['text']: attr['coding'][0]['display'] for attr in practitioner_role['resource']['characteristic']}
        practitioner_ref_to_role[practitioner_role['resource']['practitioner']['reference']] = {
            'department': practitioner_role['resource']['specialty'][0]['text'],
            'specialty': {
                'name': practitioner_role['resource']['specialty'][0]['coding'][0]['display'],
                'code': practitioner_role['resource']['specialty'][0]['coding'][0]['code']
            },
            'capacity_per_hour': int(attributes['capacity_per_hour']),
            'capacity': int(attributes['capacity']),
        }
    schedule_ref_to_practioner_ref = {
        f"Schedule/{schedule['resource']['id']}": schedule['resource']['actor'][0]['reference'] for schedule in schedules
    }

    # Append fixed schedules of a doctor
    for slot in slots:
        resource = slot['resource']
        practitioner_ref = schedule_ref_to_practioner_ref[slot['resource']['schedule']['reference']]
        date = iso_to_date(resource['start'])
        practitioner_dict = practitioner_ref_to_schedules.setdefault(practitioner_ref, {})
        practitioner_dict.setdefault(date, [])
        if not resource['status'] == 'free':
            practitioner_dict[date].append([iso_to_hour(resource['start']), iso_to_hour(resource['end'])])

    # Merge fixed schedule times
    if all(k in kwargs for k in ['start', 'end', 'interval']):
        for fixed_schedules in practitioner_ref_to_schedules.values():
            for date, time_list in fixed_schedules.items():
                fixed_schedules[date] = convert_time_list_to_merged_time(time_list=sort_schedule(time_list), **kwargs)

    # Append patient appointments of a doctor
    for appointment in appointments:
        resource = appointment['resource']
        for participant in resource['participant']:
            participant_ref = participant['actor']['reference']
            date = iso_to_date(resource['start'])
            practitioner_dict = practitioner_ref_to_schedules.setdefault(participant_ref, {})
            if participant_ref in practitioner_ref_to_name:
                practitioner_dict[date].append([iso_to_hour(resource['start']), iso_to_hour(resource['end'])])
                break
        
    # Build the doctor information from FHIR
    for practitioner in practitioners:
        resource = practitioner['resource']
        ref = f"Practitioner/{resource['id']}"
        doctor_information[practitioner_ref_to_name[ref]] = {
            'department': practitioner_ref_to_role[ref]['department'],
            'specialty': practitioner_ref_to_role[ref]['specialty'],
            'schedule': sort_schedule(practitioner_ref_to_schedules.get(ref, [])),
            'capacity_per_hour': practitioner_ref_to_role[ref]['capacity_per_hour'],
            'capacity': practitioner_ref_to_role[ref]['capacity'],
            'gender': resource['gender'],
            'telecom': resource['telecom'],
            'birthDate': resource['birthDate']
        }

    return doctor_information



import os
from importlib import resources
from typing import Optional, Tuple
from langchain_openai import ChatOpenAI
from langchain_google_genai import ChatGoogleGenerativeAI
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain.agents import (
    AgentExecutor,
    create_openai_tools_agent, 
    create_tool_calling_agent, 
)
from patientsim.utils.common_utils import set_seed

from h_adminsim.utils import colorstr, log
from h_adminsim.tools import SchedulingRule, create_tools
from h_adminsim.client import GeminiClient, GPTClient, VLLMClient



class AdminStaffAgent:
    def __init__(self,
                 target_task: str,
                 model: str,
                 api_key: Optional[str] = None,
                 use_vllm: bool = False,
                 vllm_endpoint: Optional[str] = None,
                 system_prompt_path: Optional[str] = None,
                 scheduling_user_prompt_path: Optional[str] = None,
                 tool_calling_prompt_path: Optional[str] = None,
                 sc_tool_calling_prompt_path: Optional[str] = None,
                 **kwargs):
        
        # Initialize environment
        self.target_task = target_task
        self._init_env(**kwargs)
        
        # Initialize model, API client, and other parameters
        self.model = model
        self._init_model(
            model=self.model,
            api_key=api_key,
            use_vllm=use_vllm,
            vllm_endpoint=vllm_endpoint,
        )
        
        # Initialize prompt
        self.system_prompt, self.scheduling_user_prompt_template, self.tool_calling_prompt, self.sc_tool_calling_prompt = \
            self._init_prompt(
                system_prompt_path=system_prompt_path, 
                scheduling_user_prompt_path=scheduling_user_prompt_path,
                tool_calling_prompt_path=tool_calling_prompt_path,
                sc_tool_calling_prompt_path=sc_tool_calling_prompt_path,
            )
        
        log("Administrative staff agent initialized successfully", color=True)
    

    def _init_env(self, **kwargs):
        """
        Initialize the environment with default settings.
        """
        assert self.target_task in ['first_outpatient_intake', 'first_outpatient_scheduling'], \
            log(colorstr("red", f"Unsupported target task: {self.target_task}. Supported tasks are 'first_outpatient_intake' and 'first_outpatient_scheduling'."))

        self.random_seed = kwargs.get('random_seed', None)
        self.temperature = kwargs.get('temperature', 0.2)   # For various responses. If you want deterministic responses, set it to 0.
        self.general_staff_greet = kwargs.get('general_staff_greet', "How can I help you?")
        self.staff_greet = kwargs.get('staff_greet', "How would you like to schedule the appointment?")
        self.staff_suggestion = kwargs.get('staff_suggestion', "How about this schedule: {schedule}")
        
        # Set random seed for reproducibility
        if self.random_seed:
            set_seed(self.random_seed)


    def _init_model(self,
                    model: str,
                    api_key: Optional[str] = None,
                    use_vllm: bool = False,
                    vllm_endpoint: Optional[str] = None):
        """
        Initialize the model and API client based on the specified model type.

        Args:
            model (str): The administration office agent model to use.
            api_key (Optional[str], optional): API key for the model. If not provided, it will be fetched from environment variables.
                                               Defaults to None.
            use_vllm (bool): Whether to use vLLM client.
            vllm_endpoint (Optional[str], optional): Path to the vLLM server. Defaults to None.

        Raises:
            ValueError: If the specified model is not supported.
        """
        if 'gemini' in model.lower():
            self.client = GeminiClient(model, api_key)
        elif 'gpt' in model.lower():       # TODO: Support o3, o4 models etc.
            self.client = GPTClient(model, api_key)
        elif use_vllm:
            self.client = VLLMClient(model, vllm_endpoint)
        else:
            raise ValueError(colorstr("red", f"Unsupported model: {model}. Supported models are 'gemini' and 'gpt'."))
        

    def _init_prompt(self, 
                     system_prompt_path: Optional[str] = None, 
                     scheduling_user_prompt_path: Optional[str] = None,
                     tool_calling_prompt_path: Optional[str] = None,
                     sc_tool_calling_prompt_path: Optional[str] = None) -> Tuple[str, str, str, str]:
        """
        Initialize the system prompt for the administration staff agent.

        Args:
            system_prompt_path (Optional[str], optional): Path to a custom system prompt file. 
                                                          If not provided, the default system prompt will be used. Defaults to None.
            scheduling_user_prompt_path (Optional[str], optional): Path to a custom user prompt file. 
                                                                   If not provided, the default user prompt will be used. Defaults to None.
            tool_calling_prompt_path (Optional[str], optional): Path to a custom tool calling prompt file. 
                                                                If not provided, the default tool calling prompt will be used. Defaults to None.
        Raises:
            FileNotFoundError: If the specified system prompt file does not exist.

        Returns:
            Tuple[str, str, str, str]: The system prompt, user prompt templates, tool calling prompt, and the only scheduling tool calling prompt.
        """
        # Initialilze with the default system prompt
        if not system_prompt_path:
            prompt_file_name = 'schedule_staff_system.txt'
            file_path = resources.files("h_adminsim.assets.prompts").joinpath(prompt_file_name)
            system_prompt = file_path.read_text()
        
        # User can specify a custom system prompt
        else:
            if not os.path.exists(system_prompt_path):
                raise FileNotFoundError(colorstr("red", f"System prompt file not found: {system_prompt_path}"))
            with open(system_prompt_path, 'r') as f:
                system_prompt = f.read()

        # Initialilze with the default user prompt for scheduling task
        if not scheduling_user_prompt_path:
            prompt_file_name = 'schedule_staff_reasoning.txt'
            file_path = resources.files("h_adminsim.assets.prompts").joinpath(prompt_file_name)
            scheduling_user_prompt_template = file_path.read_text()
        
        # User can specify a custom user prompt
        else:
            if not os.path.exists(scheduling_user_prompt_path):
                raise FileNotFoundError(colorstr("red", f"User prompt file not found: {scheduling_user_prompt_path}"))
            with open(scheduling_user_prompt_path, 'r') as f:
                scheduling_user_prompt_template = f.read()

        # Initialilze with the default tool calling prompt
        if not tool_calling_prompt_path:
            prompt_file_name = 'schedule_staff_tool_calling.txt'
            file_path = resources.files("h_adminsim.assets.prompts").joinpath(prompt_file_name)
            tool_calling_prompt = file_path.read_text()
        
        # User can specify a custom tool calling prompt
        else:
            if not os.path.exists(tool_calling_prompt_path):
                raise FileNotFoundError(colorstr("red", f"User prompt file not found: {tool_calling_prompt_path}"))
            else:
                with open(tool_calling_prompt_path, 'r') as f:
                    tool_calling_prompt = f.read()
        
        # Initialilze with the only scheduling tool calling prompt
        if not sc_tool_calling_prompt_path:
            prompt_file_name = 'schedule_staff_sc_tool_calling.txt'
            file_path = resources.files("h_adminsim.assets.prompts").joinpath(prompt_file_name)
            sc_tool_calling_prompt = file_path.read_text()
        
        # User can specify a custom scheduling tool calling prompt
        else:
            if not os.path.exists(sc_tool_calling_prompt_path):
                raise FileNotFoundError(colorstr("red", f"User prompt file not found: {sc_tool_calling_prompt_path}"))
            else:
                with open(sc_tool_calling_prompt_path, 'r') as f:
                    sc_tool_calling_prompt = f.read()

        return system_prompt, scheduling_user_prompt_template, tool_calling_prompt, sc_tool_calling_prompt
    

    def reset_history(self, verbose: bool = True):
        """
        Reset the conversation history.

        Args:
            verbose (bool): Whether to print verbose output. Defaults to True.
        """
        self.client.reset_history(verbose=verbose)


    def build_agent(self, 
                    rule: SchedulingRule, 
                    doctor_info: dict,
                    patient_schedule_list: Optional[list[dict]] = None,
                    gt_idx: Optional[int] = None,
                    only_schedule_tool: bool = False) -> AgentExecutor:
        """
        Build a LangChain agent with scheduling tools.

        Args:
            rule (SchedulingRule): An instance of SchedulingRule containing scheduling logic.
            doctor_info (dict): A dictionary containing information about doctors. Defaults to None.
            patient_schedule_list (Optional[list[dict]], optional): A list of the patient's scheduled appointments. Defaults to None.
            gt_idx (Optional[int], optional): Ground-truth index of the appointment to be canceled or rescheduled. Defaults to None.
            only_schedule_tool (bool, optional): Whether use only scheduling tools or not. Defaults to False.

        Returns:
            AgentExecutor: A LangChain agent executor with the scheduling tools.
        """
        tools = create_tools(rule, doctor_info, patient_schedule_list, gt_idx, only_schedule_tool)
        tool_calling_prompt = self.sc_tool_calling_prompt if only_schedule_tool else self.tool_calling_prompt
        prompt = ChatPromptTemplate.from_messages([
            ("system", tool_calling_prompt),
            MessagesPlaceholder("chat_history"),
            ("user", "{input}"),
            ("assistant", "{agent_scratchpad}"),
        ])
        # Gemini series
        if 'gemini' in self.model.lower():
            llm = ChatGoogleGenerativeAI(
                model=self.model,
                temperature=0,
            )
            agent = create_tool_calling_agent(
                llm=llm,
                tools=tools,
                prompt=prompt
            )
        # GPT series
        elif 'gpt' in self.model.lower():
            llm = ChatOpenAI(
                model_name=self.model, 
                temperature=0 if not 'gpt-5' in self.model.lower() else 1
            )
            agent = create_openai_tools_agent(
                llm=llm,
                tools=tools,
                prompt=prompt
            )
        # vLLM open sources
        else:
            llm = ChatOpenAI(
                model=self.model,
                temperature=0,
                base_url=f"{self.client.vllm_endpoint}/v1",
            )
            agent = create_openai_tools_agent(
                llm=llm,
                tools=tools,
                prompt=prompt
            )
    
        executor = AgentExecutor(
            agent=agent,
            tools=tools,
            verbose=False,
            max_iterations=1,
            return_intermediate_steps=True,
        )
        return executor
        

    def __call__(self,
                 user_prompt: str,
                 using_multi_turn: bool = True,
                 verbose: bool = True,
                 **kwargs) -> str:
        """
        Call the patient agent with a user prompt and return the response.

        Args:
            user_prompt (str): The user prompt to send to the patient agent.
            using_multi_turn (bool, optional): Whether to use multi-turn conversation. Defaults to True.
            verbose (bool, optional): Whether to print verbose output. Defaults to True.

        Returns:
            str: The response from the patient agent.
        """
        response = self.client(
            user_prompt=user_prompt,
            system_prompt=self.system_prompt,
            using_multi_turn=using_multi_turn,
            verbose=verbose,
            temperature=self.temperature,
            **kwargs
        )
        return response
    


import random
from typing import Tuple, Optional

from h_adminsim.utils.common_utils import (
    convert_segment_to_time,
    convert_time_to_segment, 
    group_consecutive_segments,
)



class ScheduleAssigner:
    def __init__(self, start: float, end: float, interval: float):
        """
        Initialize a ScheduleAssigner for generating random schedules or appointments.

        This class divides a given time range into fixed-size segments and provides methods
        to assign schedules or appointments by selecting and grouping these segments.

        Args:
            start (float): Start time in hours (e.g., 9.0 for 09:00).
            end (float): End time in hours (e.g., 18.0 for 18:00).
            interval (float): Time interval in hours for each segment (e.g., 0.5 for 30 minutes).
        """
        self.start = start
        self.end = end
        self.interval = interval
        self.segments = convert_time_to_segment(self.start, self.end, self.interval)


    def schedule_segment_assign(self,
                                p: float,
                                segments: Optional[list[int]] = None) -> list[list[int]]:
        """
        Randomly assign a proportion of schedule time segments into grouped consecutive blocks.

        This method selects a random subset of segments, where the number of segments is 
        determined by the proportion `p` (e.g., 0.5 means 50% of all segments).
        The selected segments are then grouped into lists of consecutive segment indices.

        Args:
            p (float): Proportion of total segments to assign, between 0 and 1.
            segments (Optional[list[int]], optional): Specific segments. Defaults to None.

        Returns:
            list[list[int]]: A list of groups, where each group is a list of consecutive segment indices.
                            For example, [[0, 1], [3, 4, 5], [7]].

        Example:
            If segments = [0, 1, 2, ..., 11] and p = 0.5,
            this function might return something like:
                [[0, 1], [4], [6, 7, 8]]

        Notes:
            - The segment indices are selected randomly each time the function is called.
            - Groups are always composed of consecutive indices from the selected subset.
        """
        segments = self.segments if segments == None else segments
        segment_n = round(len(segments) * p)

        if segment_n > 0:
            # Select random segments
            chosen_segments = random.sample(segments, segment_n)
            chosen_segments.sort()

            # Grouping consecutive segments
            grouped = group_consecutive_segments(chosen_segments)
            return grouped
        return []
    

    def appointment_segment_assign(self,
                                   p: float,
                                   min_chunk_size: int,
                                   max_chunk_size: int,
                                   segments: Optional[list[int]] = None) -> list[list[int]]:
        """
        Randomly assign appointment time segments from the remaining (unassigned) segments.

        Args:
            p (float): Proportion of remaining segments to sample and assign.
            min_chunk_size (int): The minimum time segment size for each appointment.
            max_chunk_size (int): The maximum time segment size for each appointment.
            segments (Optional[list[int]], optional): Specific segments. Defaults to None.

        Returns:
            list[list[int]]: Newly assigned segments from the remaining pool, grouped consecutively.
        """
        segments = self.segments if segments == None else segments
        avg_chunk_size = (min_chunk_size + max_chunk_size) / 2
        segment_n = int(len(segments) * p // avg_chunk_size)

        if segment_n > 0:
            chosen = []
            used = set()
            max_index = max(segments)
            random.shuffle(segments)
            
            for s in segments:
                max_possible_size = min(max_chunk_size, max_index - s + 1)
                if max_possible_size < min_chunk_size:
                    continue

                chunk_size = random.randint(min_chunk_size, max_chunk_size)
                trip = set(range(s, s + chunk_size))

                if used.isdisjoint(trip):
                    chosen.append((s, chunk_size))
                    used |= trip
                    if len(chosen) == segment_n:
                        break

            return [list(range(start, start + size)) for start, size in sorted(chosen)]
        return []
    

    def __call__(self,
                 p: float,
                 is_appointment: bool = False,
                 segments: Optional[list[list[int]]] = None,
                 **kwargs) -> Tuple[list[list[int]], list[list[float]]]:
        """
        Generate grouped time ranges by randomly selecting and grouping a proportion of time segments.

        This method allows the ScheduleAssigner instance to be called directly with a proportion `p`.
        It selects a subset of time segments based on `p`, groups them into consecutive segment blocks,
        and converts each group of segment indices into their corresponding time values.

        Args:
            p (float): Proportion of total segments to select, between 0 and 1.
            is_appointment (bool, optional): Whether the generated schedules are for appointments. Defaults to False.
            segments (Optional[list[list[int]]], optional): Specific segemnts. Defaults to None.

        Returns:
            list[list[float]]: A list of grouped time segments. Each group is a list of time segemnt values.
                            For example: [[2, 3], [5, 6, 7]].
            list[list[float]]: A list of grouped time ranges. Each group is a list of time values (in hours),
                            corresponding to consecutive time segments.
                            For example: [[0.0, 0.5], [2.0, 3.0]].

        Example:
            >>> assigner = ScheduleAssigner(0, 12, 0.5)
            >>> assigner(0.25)
            [[1.0, 1.5], [4.5], [6.0, 6.5]]
        """
        time_segments = self.appointment_segment_assign(p, segments=segments, **kwargs) \
            if is_appointment else self.schedule_segment_assign(p, segments)

        if len(time_segments):
            return time_segments, [list(convert_segment_to_time(self.start, self.end, self.interval, segments)) for segments in  time_segments]
        return [], []



import os
from openai import OpenAI
from dotenv import load_dotenv, find_dotenv
from typing import List, Tuple, Optional

from h_adminsim.utils import log
from h_adminsim.utils.image_preprocess_utils import *


########### For langchain integration (currently not used) ############
# from langchain_openai import ChatOpenAI
# from langchain_core.prompts import ChatPromptTemplate
# from langchain_core.output_parsers import JsonOutputParser
# from h_adminsim.registry import ScheduleModel
#######################################################################



class GPTClient:
    def __init__(self, model: str, api_key: Optional[str] = None):
        # Iniitialize
        self.model = model
        self._init_environment(api_key)
        self.histories = list()
        self.token_usages = dict()
        self.__first_turn = True


    def _init_environment(self, api_key: Optional[str] = None):
        """
        Initialize OpenAI client.

        Args:
            api_key (Optional[str]): API key for OpenAI. If not provided, it will
                                     be loaded from environment variables.
        """
        if not api_key:
            dotenv_path = find_dotenv(usecwd=True)
            load_dotenv(dotenv_path, override=True)
            api_key = os.environ.get("OPENAI_API_KEY", None)
        self.client = OpenAI(api_key=api_key)

    
    def reset_history(self, verbose: bool = True):
        """
        Reset the conversation history.

        Args:
            verbose (bool): Whether to print verbose output. Defaults to True.
        """
        self.__first_turn = True
        self.histories = list()
        self.token_usages = dict()
        if verbose:
            log('Conversation history has been reset.', color=True)

    
    def __make_payload(self,
                       user_prompt: str,
                       image_path: Optional[str] = None,
                       image_size: Optional[Tuple[int]] = None) -> List[dict]:
        """
        Create a payload for API calls to the GPT model.

        Args:
            user_prompt (str): User prompt.
            image_path (Optional[str], optional): Image path if you need to send image. Defaults to None.
            image_size (Optional[Tuple[int]], optional): Image size to be resized. Defaults to None.

        Returns:
            List[dict]: Payload including prompts and image data.
        """
        payloads = list()
        user_contents = {"role": "user", "content": []}

        # User prompts
        user_contents["content"].append(
            {"type": "text", "text": user_prompt}
        )
        if image_path:
            base64_image = encode_resize_image(image_path, image_size) if image_size else encode_image(image_path)
            extension = 'jpeg' if image_size else get_image_extension(image_path)
            user_contents["content"].append(
                {
                    "type": "image_url", 
                    "image_url": {
                        "url": f"data:image/{extension};base64,{base64_image}"
                    }
                }
            )
        
        payloads.append(user_contents)
        
        return payloads


    def __call__(self,
                 user_prompt: str,
                 system_prompt: Optional[str] = None,
                 image_path: Optional[str] = None,
                 image_size: Optional[Tuple[int]] = None,
                 using_multi_turn: bool = False,
                 verbose: bool = True,
                 **kwargs) -> str:
        """
        Sends a chat completion request to the model with optional image input and system prompt.

        Args:
            user_prompt (str): The main user prompt or query to send to the model.
            system_prompt (Optional[str], optional): An optional system-level prompt to set context or behavior. Defaults to None.
            image_path (Optional[str], optional): Path to an image file to be included in the prompt. Defaults to None.
            image_size (Optional[Tuple[int]], optional): The target image size in (width, height) format, if resizing is needed. Defaults to None.
            using_multi_turn (bool): Whether to structure it as multi-turn. Defaults to False.
            verbose (bool): Whether to print verbose output. Defaults to True.

        Raises:
            FileNotFoundError: If `image_path` is provided but the file does not exist.
            e: Any exception raised during the API call is re-raised.

        Returns:
            str: The model's response message.
        """
        if image_path and not os.path.exists(image_path):
            raise FileNotFoundError
        
        try:
            # To ensure empty history
            if not using_multi_turn:
                self.reset_history(verbose)
            
            if self.__first_turn:
                # System prompt
                if system_prompt:
                    self.histories.append({"role": "system", "content": [{"type": "text", "text": system_prompt}]})
                self.__first_turn = False
            
            # User prompt
            self.histories += self.__make_payload(user_prompt, image_path, image_size)

            # Model response
            response = self.client.chat.completions.create(
                model=self.model,
                messages=self.histories,
                **kwargs
            )
            assistant_msg = response.choices[0].message
            self.histories.append({"role": assistant_msg.role, "content": [{"type": "text", "text": assistant_msg.content}]})

            # Logging token usage
            if response.usage:
                self.token_usages.setdefault("prompt_tokens", []).append(response.usage.prompt_tokens)
                self.token_usages.setdefault("completion_tokens", []).append(response.usage.completion_tokens)
                self.token_usages.setdefault("total_tokens", []).append(response.usage.total_tokens)
                self.token_usages.setdefault("reasoning_tokens", []).append(response.usage.completion_tokens_details.reasoning_tokens)

            return assistant_msg.content
        
        except Exception as e:
            raise e



# class GPTLangChainClient(GPTClient):
#     def __init__(self, model: str):
#         super(GPTLangChainClient, self).__init__(model)
#         self.client_lc = ChatOpenAI(
#             model=self.model,
#             api_key=self.client.api_key
#         )

    
#     def __call__(self,
#                  user_prompt: str,
#                  system_prompt: Optional[str] = None,
#                  image_path: Optional[str] = None,
#                  image_size: Optional[Tuple[int]] = None,
#                  using_multi_turn: bool = False,
#                  **kwargs) -> str:
#         try:
#             # To ensure empty history
#             self.reset_history()

#             # Prompts
#             parser = JsonOutputParser(pydantic_object=ScheduleModel)
#             prompt = ChatPromptTemplate.from_messages(
#                 [
#                     ('system', system_prompt),
#                     ('human', user_prompt)
#                 ]
#             ).partial(format_instructions=parser.get_format_instructions())
#             chain = prompt | self.client_lc | parser
            
#             # Model response
#             response = chain.invoke(kwargs)

#             return response
        
#         except Exception as e:
#             raise e



from .scheduling_rule import *
from .data_converter import DataConverter
from .agent_data_builder import AgentDataBuilder
from .data_synthesizer import DataSynthesizer
from .evaluator import Evaluator


import os
import json
import random
from copy import deepcopy
from decimal import getcontext
from importlib import resources
from typing import Tuple, Union, Optional
from dotenv import load_dotenv, find_dotenv

from patientsim import PatientAgent
from patientsim import AdminStaffAgent as IntakeAdminStaffAgent
from patientsim.environment import OPSimulation as OPFVIntakeSimulation

from h_adminsim import SupervisorAgent
from h_adminsim import AdminStaffAgent as SchedulingAdminStaffAgent
from h_adminsim.environment.hospital import HospitalEnvironment
from h_adminsim.environment import OPScehdulingSimulation as OPFVScheduleSimulation
from h_adminsim.tools.sanity_checker import SanityChecker
from h_adminsim.tools import DataConverter, SchedulingRule
from h_adminsim.registry import STATUS_CODES, PREFERENCE_PHRASE_PATIENT
from h_adminsim.utils import colorstr, log
from h_adminsim.utils.fhir_utils import *
from h_adminsim.utils.common_utils import *



class FirstVisitOutpatientTask:
    def __init__(self):
        self.token_stats = {
            'patient_token': {'input':[], 'output': [], 'reasoning': []}, 
            'admin_staff_token': {'input': [], 'output': [], 'reasoning': []}, 
            'supervisor_token': {'input':[], 'output': [], 'reasoning': []}
        }

    
    def save_token_data(self, 
                        patient_token: Optional[dict] = None, 
                        admin_staff_token: Optional[dict] = None, 
                        supervisor_token: Optional[dict] = None):
        """
        Save the API token usage data

        Args:
            patient_token (Optional[dict], optional): Patient token information. Defaults to None.
            admin_staff_token (Optional[dict], optional): Administration staff token information. Defaults to None.
            supervisor_token (Optional[dict], optional): Supervisor token information. Defaults to None.
        """
        if patient_token:
            self.token_stats['patient_token']['input'].extend(patient_token['prompt_tokens'])
            self.token_stats['patient_token']['output'].extend(patient_token['completion_tokens'])
            if 'reasoning_tokens' in patient_token:
                self.token_stats['patient_token']['reasoning'].extend(patient_token['reasoning_tokens'])

        if admin_staff_token:
            self.token_stats['admin_staff_token']['input'].extend(admin_staff_token['prompt_tokens'])
            self.token_stats['admin_staff_token']['output'].extend(admin_staff_token['completion_tokens'])
            if 'reasoning_tokens' in admin_staff_token:
                self.token_stats['admin_staff_token']['reasoning'].extend(admin_staff_token['reasoning_tokens'])

        if supervisor_token:
            self.token_stats['supervisor_token']['input'].extend(supervisor_token['prompt_tokens'])
            self.token_stats['supervisor_token']['output'].extend(supervisor_token['completion_tokens'])
            if 'reasoning_tokens' in supervisor_token:
                self.token_stats['supervisor_token']['reasoning'].extend(supervisor_token['reasoning_tokens'])
    
    
    def _init_task_models(self, model: str, vllm_endpoint: Optional[str] = None) -> Tuple[str, str, bool]:
        """
        Initialize the model for the task.

        Args:
            model (str): The model name.
            vllm_endpoint (Optional[str], optional): The VLLM endpoint URL. Defaults to None.
        
        Returns:
            Tuple[str, str, bool]: The model name, VLLM endpoint URL, vllm usage flag.
        """
        if any(keyword in model.lower() for keyword in ['gemini', 'gpt']):
            return model, None, False
        else:
            assert vllm_endpoint is not None, log('VLLM endpoint must be provided for non-Gemini/GPT models.', 'error')
            return model, vllm_endpoint, True



class OutpatientFirstIntake(FirstVisitOutpatientTask):
    def __init__(self, 
                 patient_model: str,
                 admin_staff_model: str,
                 supervisor_agent: Optional[SupervisorAgent] = None,
                 intake_max_inference: int = 5,
                 max_retries: int = 8,
                 admin_staff_last_task_user_prompt_path: Optional[str] = None,
                 patient_vllm_endpoint: Optional[str] = None,
                 admin_staff_vllm_endpoint: Optional[str] = None):
        super().__init__()
        
        # Initialize variables
        self.name = 'intake'
        self.patient_model, self.patient_vllm_endpoint, self.patient_use_vllm \
            = self._init_task_models(patient_model, patient_vllm_endpoint)
        self.admin_staff_model, self.admin_staff_vllm_endpoint, self.admin_staff_use_vllm \
            = self._init_task_models(admin_staff_model, admin_staff_vllm_endpoint)
        self.use_supervisor = True if isinstance(supervisor_agent, SupervisorAgent) else False
        self.supervisor_client = supervisor_agent if self.use_supervisor else None
        task_mechanism = 'Staff + Supervisor' if self.use_supervisor else 'Staff'
        self.max_inferences = intake_max_inference
        self.max_retries = max_retries
        self._init_last_task_prompt(admin_staff_last_task_user_prompt_path)
        self.patient_reasoning_kwargs = {'reasoning_effort': 'low'} if 'gpt-5' in self.patient_model.lower() else {}
        self.staff_reasoning_kwargs = {'reasoning_effort': 'low'} if 'gpt-5' in self.admin_staff_model.lower() else {}
        log(f'Patient intake tasks are conducted by {colorstr(task_mechanism)}')
    

    def _init_last_task_prompt(self, admin_staff_last_task_user_prompt_path: Optional[str] = None) -> str:
        """
        Initialize the user prompt for the admnistration staff agent's last task.

        Args:
            admin_staff_last_task_user_prompt_path (Optional[str], optional): Path to a custom user prompt file. 
                                                                              If not provided, the default user prompt will be used. Defaults to None.

        Raises:
            FileNotFoundError: If the specified user prompt file does not exist.
            
        Returns:
            str: The user prompt.
        """
        if not self.use_supervisor:
            if not admin_staff_last_task_user_prompt_path:
                prompt_file_name = 'intake_staff_task_user.txt'
                file_path = resources.files("h_adminsim.assets.prompts").joinpath(prompt_file_name)
                self.last_task_user_prompt = file_path.read_text()
            else:
                if not os.path.exists(admin_staff_last_task_user_prompt_path):
                    raise FileNotFoundError(colorstr("red", f"User prompt file not found: {admin_staff_last_task_user_prompt_path}"))
                with open(admin_staff_last_task_user_prompt_path, 'r') as f:
                    self.last_task_user_prompt = f.read()
        else:
            if admin_staff_last_task_user_prompt_path:
                log('The admin_staff_last_task_user_prompt_path setting is ignored when using supervisor model.', 'warning')
    
    
    @staticmethod
    def postprocessing_department(text: str) -> str:
        """
        Post-processing method of text output, especially for the department decision.

        Args:
            text (str): Text input.

        Returns:
            str: Post-processed text output.
        """
        try:
            pattern = re.compile(r'Answer:\s*\d+\.\s*(.+)')
            text = pattern.search(text).group(1)
        except:
            text = 'wrong'
        return text

    
    @staticmethod
    def postprocessing_information(text: str) -> Union[str, dict]:
        """
        Post-processing method of text output, especially for the patient information extraction.

        Args:
            text (str): Text input.

        Returns:
            Union[str, dict]: A dictionary if the text is valid JSON, otherwise the original string.
        """
        try:
            if isinstance(text, str):
                match = re.search(r'```json\s*(\{.*?\})\s*```', text, re.DOTALL)
                if match:
                    json_str = match.group(1)
                    text_dict = json.loads(json_str)
                else:
                    try:
                        text_dict = json.loads(text)
                    except:
                        return text
            else:
                text_dict = text
            
            assert len(text_dict) == 6 and all(k in text_dict for k in ['name', 'gender', 'phone_number', 'personal_id', 'address', 'department'])   # Basic sanity check
            return text_dict
        except:
            return str(text)
        
    
    def _department_decision(self, prediction_department: str, prediction_supervison: Union[str, dict], gt_department: str) -> Tuple[str, list[str]]:
        """
        Determine the final department decision by considering both 
        the interaction agent result and the supervisor agent result.

        Args:
            prediction_department (str): The department predicted by the interaction agent.
            prediction_supervison (Union[str, dict]): The supervisor agent's result. 
                If this is a dictionary, it should contain a 'department' field.
            gt_department (str): The ground truth department for the patient.

        Returns:
            str: The final department decision.
        """
        try:
            sup_department = prediction_supervison.pop('department')
            if prediction_department == sup_department:
                trial = ['match']
            else:
                if prediction_department in gt_department and sup_department not in gt_department:
                    trial = ['mismatch - worse']
                elif prediction_department not in gt_department and sup_department in gt_department:
                    trial = ['mismatch - better']
                else:
                    trial = ['mismatch - both wrong']
            return sup_department, trial
        except:
            return prediction_department, ['supervisor error']
        

    def __call__(self, data_pair: Tuple[dict, dict], agent_test_data: dict, agent_results: dict, environment, verbose: bool = False) -> dict:
        """
        Estimates the most appropriate medical department for each patient using an LLM agent.

        Args:
            data_pair (Tuple[dict, dict]): A pair of ground truth and patient data for agent simulation.
            agent_test_data (dict): A dictionary containing test data for a single hospital.
                Expected to include:
                    - 'department': Dictionary of available departments.
            agent_results (dict): Placeholder for compatibility; not used in this method.
            environment (HospitalEnvironment): Hospital environment instance to manage patient schedules.
            verbose (bool): Whether logging the each result or not.

        Returns:
            dict: A dictionary with:
                - 'gt': List of ground-truth departments.
                - 'pred': List of predicted departments from the LLM agent.
                - 'status': List of booleans indicating whether each prediction correct.
                - 'status_code': List of status codes explaining each status.
        """
        gt, test_data = data_pair
        departments = list(agent_test_data['department'].keys())
        results = init_result_dict()
        sanity_checker = SanityChecker()
        
        # Append a ground truth
        name, gender, birth_date, telecom, personal_id, address = \
            gt['patient'], gt['gender'], gt['birthDate'], gt['telecom'][0]['value'], gt['identifier'][0]['value'], gt['address'][0]['text']
        gt_data = {
            'patient': {
                'name': name,
                'gender': gender,
                'phone_number': telecom,
                'personal_id': personal_id,
                'address': address,
            }, 
            'department': gt['department']
        }
        results['gt'].append(gt_data)

        # LLM call: Conversation and department decision
        department_candidates = test_data['constraint']['symptom']['department']
        if test_data['constraint']['symptom_level'] == 'simple':
            medical_history = "None. This is the patient's first visit."
            diagnosis = "Unknown for now, as this is the patient's first visit to the hospital."
        elif test_data['constraint']['symptom_level'] == 'with_history':
            medical_history = f"Diagnosed with {test_data['constraint']['symptom']['disease']} at a primary or secondary hospital."
            diagnosis = test_data['constraint']['symptom']['disease']
        else:
            log("Patient's symptom level must be either 'simple' or 'with_history'.", "error")
        
        # Simulation patient intake
        patient_agent = PatientAgent(
            self.patient_model,
            'outpatient',
            lang_proficiency_level='B',
            recall_level='no_history' if test_data['constraint']['symptom_level'] == 'simple' else 'high',
            use_vllm=self.patient_use_vllm,
            vllm_endpoint=self.patient_vllm_endpoint,
            department=department_candidates,
            name=name,
            birth_date=birth_date,
            gender=gender,
            telecom=telecom,
            personal_id=personal_id,
            address=address,
            medical_history=medical_history,
            diagnosis=diagnosis,
            chiefcomplaint=test_data['constraint']['symptom']['symptom'],
            temperature=0 if not 'gpt-5' in self.patient_model.lower() else 1
        )
        admin_staff_agent = IntakeAdminStaffAgent(
            self.admin_staff_model,
            departments,
            max_inferences=self.max_inferences,
            use_vllm=self.admin_staff_use_vllm,
            vllm_endpoint=self.admin_staff_vllm_endpoint,
            temperature=0 if not 'gpt-5' in self.admin_staff_model.lower() else 1
        )
        sim_environment = OPFVIntakeSimulation(patient_agent, admin_staff_agent, max_inferences=self.max_inferences)
        output = run_with_retry(
            sim_environment.simulate,
            verbose=False,
            patient_kwargs=self.patient_reasoning_kwargs, 
            staff_kwargs=self.staff_reasoning_kwargs,
            max_retries=self.max_retries,
        )
        dialogs, patient_token, admin_staff_token = output['dialog_history'], output.get('patient_token_usage'), output.get('admin_staff_token_usage')
        prediction_department = OutpatientFirstIntake.postprocessing_department(dialogs[-1]['content'])

        # LLM call: Agent which should extract demographic information of the patient and evaluation the department decision result
        dialogs = preprocess_dialog(dialogs)
        
        if self.use_supervisor:
            user_prompt = self.supervisor_client.user_prompt_template.format(
                CONVERSATION=dialogs,
                DEPARTMENTS=''.join([f'{i+1}. {department}\n' for i, department in enumerate(departments)])
            )
            prediction_supervision = run_with_retry(
                self.supervisor_client,
                user_prompt,
                using_multi_turn=False,
                verbose=False,
                max_retries=self.max_retries,
            )
        else:
            prediction_supervision = run_with_retry(
                admin_staff_agent,
                self.last_task_user_prompt,
                verbose=False,
                max_retries=self.max_retries,
                **self.staff_reasoning_kwargs,
            )

        prediction_supervision = OutpatientFirstIntake.postprocessing_information(prediction_supervision)
        
        # Append token data
        self.save_token_data(
            patient_token, 
            admin_staff_token, 
            supervisor_token=self.supervisor_client.client.token_usages if self.use_supervisor else {}
        )

        # Sanity check
        department, trial = self._department_decision(prediction_department, prediction_supervision, gt['department'])
        prediction = {'patient': prediction_supervision, 'department': [department]}
        status, status_code = sanity_checker.intake_check(
            prediction=prediction,
            gt=gt_data,
            conversations=dialogs
        )

        if verbose:
            log(f'GT    : {gt_data}')
            log(f'Pred  : {prediction}')
            log(f'Status: {status_code}\n\n\n')
        
        # Append results
        results['pred'].append(prediction)
        results['status'].append(status)
        results['status_code'].append(status_code)
        results['trial'].append(trial)
        results['dialog'].append(dialogs)

        return results



class OutpatientFirstScheduling(FirstVisitOutpatientTask):
    def __init__(self, 
                 patient_model: str,
                 admin_staff_model: str,
                 schedule_cancellation_prob: float = 0.05,
                 request_early_schedule_prob: float = 0.1,
                 preference_rejection_prob: float = 0.3,
                 preference_rejection_prob_decay: float = 0.5,
                 fhir_integration: bool = False,
                 scheduling_max_inference: int = 5,
                 scheduling_strategy: str = 'tool_calling',
                 max_retries: int = 8,
                 patient_vllm_endpoint: Optional[str] = None,
                 admin_staff_vllm_endpoint: Optional[str] = None):
        super().__init__()

        # Initialize variables
        getcontext().prec = 10
        dotenv_path = find_dotenv(usecwd=True)
        load_dotenv(dotenv_path, override=True)
        self.name = 'schedule'
        self.patient_model, self.patient_vllm_endpoint, self.patient_use_vllm \
            = self._init_task_models(patient_model, patient_vllm_endpoint)
        self.admin_staff_model, self.admin_staff_vllm_endpoint, self.admin_staff_use_vllm \
            = self._init_task_models(admin_staff_model, admin_staff_vllm_endpoint)
        
        # Initialize scheduling methods and a staff agent
        self.admin_staff_agent = SchedulingAdminStaffAgent(
            target_task='first_outpatient_scheduling',
            model=self.admin_staff_model,
            use_vllm=self.admin_staff_use_vllm,
            vllm_endpoint=self.admin_staff_vllm_endpoint,
            temperature=0 if not 'gpt-5' in self.admin_staff_model.lower() else 1
        )

        # Scheduling parameters
        self.schedule_cancellation_prob = schedule_cancellation_prob
        self.request_early_schedule_prob = request_early_schedule_prob
        self.preference_rejection_prob = preference_rejection_prob
        self.preference_rejection_prob_decay = preference_rejection_prob_decay

        # Others
        self.fhir_integration = fhir_integration
        self.max_retries = max_retries
        self.max_inferences = scheduling_max_inference
        self.scheduling_strategy = scheduling_strategy
        assert self.scheduling_strategy in ['reasoning', 'tool_calling'], \
            log('Scheduling strategy must be either `reasoning` or `tool_calling`.', 'error')
        self.schedule_patient_system_prompt_path = str(resources.files("h_adminsim.assets.prompts").joinpath('schedule_patient_system.txt'))
        self.cancel_patient_system_prompt_path = str(resources.files("h_adminsim.assets.prompts").joinpath('cancel_patient_system.txt'))
        self.reschedule_patient_system_prompt_path = str(resources.files("h_adminsim.assets.prompts").joinpath('reschedule_patient_system.txt'))
        self.patient_reasoning_kwargs = {'reasoning_effort': 'low'} if 'gpt-5' in self.patient_model.lower() else {}
        self.staff_reasoning_kwargs = {'reasoning_effort': 'low'} if 'gpt-5' in self.admin_staff_model.lower() else {}

    
    def _init_simulation(self,
                         system_prompt_path: str,
                         environment: HospitalEnvironment,
                         additional_patient_conditions: dict = {}) -> OPFVScheduleSimulation:
        """
        Initialize an outpatient first-visit intake and scheduling simulation.

        Args:
            system_prompt_path (str): Path to the system prompt used to initialize the patient agent.
            environment (HospitalEnvironment): Hospital environment configuration for the simulation.
            additional_patient_conditions (dict, optional): Additional patient-specific conditions for simulation control.

        Returns:
            OPFVIntakeSimulation: Configured outpatient intake and scheduling simulation instance.
        """
        patient_agent = PatientAgent(
            self.patient_model,
            'outpatient',
            use_vllm=self.patient_use_vllm,
            vllm_endpoint=self.patient_vllm_endpoint,
            system_prompt_path=system_prompt_path,
            log_verbose=False,
            additional_patient_conditions=additional_patient_conditions,
            temperature=0 if not 'gpt-5' in self.patient_model.lower() else 1
        )
        sim_environment = OPFVScheduleSimulation(
            patient_agent=patient_agent, 
            admin_staff_agent=self.admin_staff_agent, 
            metadata=self._metadata,
            department_data=self._department_data,
            environment=environment,
            scheduling_strategy=self.scheduling_strategy,
            preference_rejection_prob=self.preference_rejection_prob,
            preference_rejection_prob_decay=self.preference_rejection_prob_decay,
            fhir_integration=self.fhir_integration,
            sanity_checker=self.sanity_checker, 
        )
        return sim_environment


    def get_intake_information(self, gt: dict, agent_results: dict, doctor_information: dict) -> Tuple[dict, str, bool]:
        """
        Extracts the patient name and predicted department from agent results.
        If predictions are not available, falls back to using ground truth labels.

        Args:
            gt (dict): Ground truth data of a patient.
            agent_results (dict): A dictionary that may contain predicted department results under the key 'department'.
            doctor_information (dict): Dictionary of doctor data including their existing schedules.
                                       Each key is a doctor's name, and each value includes a 'schedule' field.

        Returns:
            Tuple[dict, str, bool]: Patient information, determined department, either predicted or ground truth and its sanity status.
        """
        # Prediction results are existing case
        try:
            for i, intake_gt in enumerate(agent_results['intake']['gt']):
                if gt['patient'] == intake_gt['patient']['name']:
                    break

            patient_info = agent_results['intake']['pred'][i]['patient']
            department = agent_results['intake']['pred'][i]['department'][0]
            sanity = agent_results['intake']['status'][i]

            assert gt['patient'] == agent_results['intake']['gt'][i]['patient']['name']
        
        # Loading from the ground truth
        except:
            log('The predicted department is not given. The ground truth value will be used.', 'warning')
            patient_info = {
                'name': gt['patient'],
                'gender': gt['gender'],
                'phone_number': gt['telecom'][0]['value'],
                'personal_id': gt['identifier'][0]['value'],
                'address': gt['address'][0]['text'],
            }
            department = doctor_information[gt['attending_physician']]['department']
            assert department in gt['department']
            sanity = True

        return patient_info, department, sanity


    def cancellation_request(self, 
                        doctor_information: dict, 
                        environment: HospitalEnvironment, 
                        idx: Optional[int] = None, 
                        verbose: bool = False) -> Tuple[dict, Optional[dict]]:
        """
        Cancel a doctor's scheduled appointment.

        Args:
            doctor_information (dict): A dictionary containing information about the doctor(s) involved,
                                       including availability and other relevant details.
            environment (HospitalEnvironment): Hospital environment.
            idx (int, optional): Specific patient schedule index.
            verbose (bool, optional): Whether logging the each result or not. Defaults to False.

        Returns:
            Tuple[dict, Optional[dict]]: Updated doctor information and a result dictionary after cancellation.
        """
        if idx is None:
            candidate_idx = [i for i, schedule in enumerate(environment.patient_schedules) if schedule['status'] == 'scheduled']
            idx = random.choice(candidate_idx) if len(candidate_idx) else -1

        if idx >= 0:
            # Ground-truth cancelled schedule
            cancelled_schedule = environment.patient_schedules[idx]
            patient = cancelled_schedule['patient']
            doctor, date, time = cancelled_schedule['attending_physician'], cancelled_schedule['date'], cancelled_schedule['schedule']
            
            # Initialize simulation environment for cancellation
            sim_environment = self._init_simulation(
                system_prompt_path=self.cancel_patient_system_prompt_path,
                environment=environment,
                additional_patient_conditions={
                    'patient_name': patient,
                    'doctor_name': doctor,
                    'date': date,
                    'start_time': hour_to_hhmmss(time[0])
                }
            )

            # Schedule cancellation simulation
            doctor_information, result_dict = run_with_retry(
                sim_environment.canceling_simulate,
                gt_idx=idx,
                doctor_information=doctor_information,
                patient_schedules=environment.patient_schedules,
                verbose=verbose,
                max_inferences=self.max_inferences,
                patient_kwargs=self.patient_reasoning_kwargs,
                staff_kwargs=self.staff_reasoning_kwargs,
                max_retries=self.max_retries,
            )

            # Successfully canceled
            if result_dict['status'][0] is not False:   # No GT and correct case
                # Update waiting list due to cancellation
                doctor_information, rs_result_dict = self.automatic_waiting_list_update(
                    sim_environment=sim_environment,
                    environment=environment,
                    doctor_information=doctor_information,
                )

                # Update result dictionary
                for key in result_dict.keys():
                    if len(rs_result_dict[key]):
                        result_dict[key].append(tuple(rs_result_dict[key]))
            
            return doctor_information, result_dict

        return doctor_information, None
                

    def rescheduling_request(self,
                             doctor_information: dict,
                             environment: HospitalEnvironment, 
                             idx: Optional[int] = None, 
                             verbose: bool = False) -> Tuple[dict, Optional[dict]]:
        """
        Add a patient schedule to the waiting list in the given environment.

        Args:
            doctor_information (dict): A dictionary containing information about the doctor(s) involved,
                                       including availability and other relevant details.
            environment (HospitalEnvironment): Hospital environment.
            idx (int, optional): Specific patient schedule index.
            verbose (bool, optional): Whether logging the each result or not. Defaults to False.
        
        Returns:
            Tuple[dict, Optional[dict]]: Updated doctor information and a result dictionary after cancellation.
        """
        result_dict = init_result_dict()
        if idx is None:
            candidate_idx = [i for i, schedule in enumerate(environment.patient_schedules) if schedule['status'] == 'scheduled']
            idx = random.choice(candidate_idx) if len(candidate_idx) else -1
        
        if idx >= 0:
            requested_schedule = environment.patient_schedules[idx]
            if all(requested_schedule != s[1] for s in environment.waiting_list):
                # Ground-truth rescheduling requested schedule
                patient = requested_schedule['patient']
                doctor, date, time = requested_schedule['attending_physician'], requested_schedule['date'], requested_schedule['schedule']

                # Initialize simulation environment for rescheduling request
                sim_environment = self._init_simulation(
                    system_prompt_path=self.reschedule_patient_system_prompt_path,
                    environment=environment,
                    additional_patient_conditions={
                        'patient_name': patient,
                        'doctor_name': doctor,
                        'date': date,
                        'start_time': hour_to_hhmmss(time[0])
                    }
                )

                # Rescheduling request simulation
                doctor_information, result_dict = run_with_retry(
                    sim_environment.rescheduling_simulate,
                    gt_idx=idx,
                    doctor_information=doctor_information,
                    patient_schedules=environment.patient_schedules,
                    verbose=verbose,
                    max_inferences=self.max_inferences,
                    patient_kwargs=self.patient_reasoning_kwargs,
                    staff_kwargs=self.staff_reasoning_kwargs,
                    max_retries=self.max_retries,
                )

                if result_dict['status'][0] is not False:   # No GT and correct case
                    if 'patient' in result_dict['pred'][0]:
                        new_schedule = result_dict['pred'][0]
                        doctor_information[new_schedule['attending_physician']]['schedule'][new_schedule['date']].append(new_schedule['schedule'])
                        doctor_information[new_schedule['attending_physician']]['schedule'][new_schedule['date']].sort()
                        self.update_env(
                            status=True,
                            prediction=new_schedule,
                            environment=environment,
                        )

                return doctor_information, result_dict

            return doctor_information, None

        return doctor_information, None
            

    def automatic_waiting_list_update(self, 
                                      sim_environment: OPFVScheduleSimulation,
                                      environment: HospitalEnvironment,
                                      doctor_information: Optional[dict] = None) -> Tuple[dict, dict]:
        """
        Automatically update the waiting list by attempting to reschedule patients.

        Args:
            sim_environment (OPFVScheduleSimulation): The simulation environment used for scheduling.
            environment (HospitalEnvironment): Hospital environment.
            doctor_information (Optional[dict], optional): A dictionary containing information about the doctor(s) involved, 
                                                           including availability and other relevant details. Defaults to None.

        Returns:
            Tuple[dict, dict]: Updated doctor information and a result dictionary.
        """
        all_result_dict = init_result_dict()
        for result in sim_environment.automatic_waiting_list_update(
            doctor_information=doctor_information,
            **self.staff_reasoning_kwargs,
        ):
            doctor_information, result_dict = result['doctor_information'], result['result_dict']

            if result_dict['status'][0]:
                new_schedule, original = result_dict['pred'][0], result['original']
                doctor_information[new_schedule['attending_physician']]['schedule'][new_schedule['date']].append(new_schedule['schedule'])
                doctor_information[new_schedule['attending_physician']]['schedule'][new_schedule['date']].sort()
                self.update_env(
                    status=True,
                    prediction=new_schedule,
                    environment=environment,
                )
                log(f'{colorstr("[RESCHEDULED]")}: {original} is rescheduled to {new_schedule}')

            all_result_dict['gt'].extend(result_dict['gt'])
            all_result_dict['pred'].extend(result_dict['pred'])
            all_result_dict['status'].extend(result_dict['status'])
            all_result_dict['status_code'].extend(result_dict['status_code'])
            all_result_dict['dialog'].extend(result_dict['dialog'])

        return doctor_information, all_result_dict
                   

    def update_env(self, 
                   status: bool, 
                   prediction: Union[dict, str], 
                   environment: HospitalEnvironment, 
                   patient_information: Optional[dict] = None):
        """
        Update the simulation environment with scheduling results and optionally synchronize FHIR resources.

        Args:
            status (bool): Whether the scheduling task was successful. If True, FHIR resources may be updated.
            prediction (Union[dict, str]): The predicted scheduling result (e.g., patient schedule information).
            environment (HospitalEnvironment): The environment instance to be updated (must implement `update_env`).
            patient_information (Optional[dict], optional): Patient-related predicted (or GT) information to generate FHIR Patient resources. Defaults to None.

        """
        # POST/PUT to FHIR
        fhir_patient, fhir_appointment = None, None
        if status and self.fhir_integration:
            if patient_information is not None:
                fhir_patient = DataConverter.data_to_patient(
                    {
                        'metadata': deepcopy(self._metadata),
                        'department': deepcopy(self._department_data),
                        'patient': {
                            prediction['patient']: {
                                'department': prediction['department'], 
                                'gender': patient_information['gender'],
                                'telecom': [{'system': 'phone', 'value': patient_information['phone_number'], 'use': 'mobile'}],
                                'birthDate': personal_id_to_birth_date(patient_information['personal_id']),
                                'identifier': [{'value': patient_information['personal_id'], 'use': 'official'}],
                                'address': [{'type': 'postal', 'text': patient_information['address'], 'use': 'home'}],
                            }
                        }
                    }
                )[0]
            fhir_appointment = DataConverter.get_fhir_appointment(data={'metadata': deepcopy(self._metadata),
                                                                        'department': deepcopy(self._department_data),
                                                                        'information': deepcopy(prediction)})
            
        environment.update_env(
            status=status, 
            patient_schedule=prediction,
            fhir_resources={'Patient': fhir_patient, 'Appointment': fhir_appointment}
        )
            

    def __call__(self, data_pair: Tuple[dict, dict], agent_test_data: dict, agent_results: dict, environment, verbose: bool = False) -> dict:
        """
        This method uses agent test data to prompt an LLM for scheduling decisions, post-processes
        the output, runs sanity checks on predicted schedules, and collects the results for evaluation.

        Args:
            data_pair (Tuple[dict, dict]): A pair of ground truth and patient data for agent simulation.
            agent_test_data (dict): Dictionary containing test data and metadata for a single hospital.
                Expected keys include:
                    - 'metadata': A dict containing start_hour, end_hour, and interval_hour under 'time'.
                    - 'agent_data': A list of (ground_truth, test_data) pairs.
                    - 'doctor': A dictionary of doctor profiles with department and schedule info.
            agent_results (dict): Optional dictionary containing prior department predictions.
                Used to extract department-level guidance per patient. Can be empty.
            environment (HospitalEnvironment): Hospital environment instance to manage patient schedules.
            verbose (bool, option): Whether logging the each result or not.

        Returns:
            dict: A dictionary with three keys:
                - 'gt': List of ground truth results, each including patient info, attending physician, department, and schedule.
                - 'pred': List of predicted results (either valid dict or fallback string).
                - 'status': List of booleans indicating whether each prediction passed sanity checks.
                - 'status_code': List of status codes explaining each status.
        """
        gt, test_data = data_pair
        self._metadata = agent_test_data.get('metadata')
        self._department_data = agent_test_data.get('department')
        self._START_HOUR = self._metadata.get('time').get('start_hour')
        self._END_HOUR = self._metadata.get('time').get('end_hour')
        self._TIME_UNIT = self._metadata.get('time').get('interval_hour')
        self.sanity_checker = SanityChecker(self._START_HOUR, self._END_HOUR, self._TIME_UNIT)
        doctor_information = environment.get_general_doctor_info_from_fhir() if self.fhir_integration else agent_test_data.get('doctor')
        patient_info, department, sanity = self.get_intake_information(gt, agent_results, doctor_information)
        self.rules = SchedulingRule(self._metadata, self._department_data, environment, self.fhir_integration)
        results = init_result_dict()

        # Make scheduling GT list
        gt_data = [
            {
                'patient': patient_info['name'] if sanity else gt.get('patient'),
                'department': department if sanity else gt.get('department'),
                'preference': preference,
                'preferred_doctor': gt.get('attending_physician') if preference == 'doctor' else "Doesn't matter",
                'valid_from': gt.get('valid_from') if preference == 'date' else None,
            } for preference in gt.get('preference')
        ]
        staff_known_data = [
            {
                'patient': patient_info['name'],
                'department': department,
                'patient_intention': None,
            } for _ in range(len(gt_data))
        ]

        # If the precedent department data is wrong, continue
        if not sanity:
            results['gt'].append(gt_data)
            results['pred'].append({})
            results['status'].append(False)
            results['status_code'].append(STATUS_CODES['preceding'])
            return results
        
        #################################################### Regular Scheudling Simulation ####################################################
        # Initialize the simulation environment using the first preference data
        preference = gt_data[0].get('preference')
        preference_desc = PREFERENCE_PHRASE_PATIENT[preference] if preference != 'date' \
                    else PREFERENCE_PHRASE_PATIENT[preference].format(date=gt_data[0].get('valid_from'))
        sim_environment = self._init_simulation(
            system_prompt_path=self.schedule_patient_system_prompt_path,
            environment=environment,
            additional_patient_conditions={
                'preference': preference,
                'preference_desc': preference_desc,
                'preferred_doctor': gt_data[0]['preferred_doctor'],
            }
        )
    
        # Simulate the main scheduling task
        doctor_information, result_dict = run_with_retry(
            sim_environment.scheduling_simulate,
            gt_data=gt_data,
            staff_known_data=staff_known_data,
            doctor_information=doctor_information,
            verbose=verbose,
            patient_kwargs=self.patient_reasoning_kwargs,
            staff_kwargs=self.staff_reasoning_kwargs,
            max_retries=self.max_retries,
        )

        prediction, status, status_code = \
            result_dict['pred'][0], result_dict['status'][0], result_dict['status_code'][0]
        
        if verbose:
            log(f'Pred  : {prediction}')
            log(f'Status: {status_code}')
            log(f'Final Status: {status_code}\n\n\n')

        # Update the simulation environment and the doctor information in the agent test data
        if status:
            doctor_information[prediction['attending_physician']]['schedule'][prediction['date']].append(prediction['schedule'])
            doctor_information[prediction['attending_physician']]['schedule'][prediction['date']].sort()
        
        self.update_env(
            status=status,
            prediction=prediction,
            environment=environment,
            patient_information=patient_info,
        )
        agent_test_data['doctor'] = doctor_information

        # Append results
        for key in result_dict.keys():
            results[key] += result_dict[key]
        #######################################################################################################################################
        
        # Other events
        ## Simulate the schedule cancellation requests
        if random.random() < self.schedule_cancellation_prob:
            doctor_information, result_dict = self.cancellation_request(
                doctor_information=doctor_information,
                environment=environment,
                verbose=verbose,
            )
            if result_dict is not None:
                agent_test_data['doctor'] = doctor_information
                results['gt'].extend(result_dict['gt'])
                results['pred'].extend(result_dict['pred'])
                results['status'].extend(result_dict['status'])
                results['status_code'].extend(result_dict['status_code'])
                results['dialog'].extend(result_dict['dialog'])

                if verbose:
                    log(f'Pred  : {result_dict["pred"]}')
                    log(f'Status: {result_dict["status_code"]}')
                    log(f'Final Status: {result_dict["status_code"]}\n\n\n')
        
        ## Simulate the resecheduling requests
        if random.random() < self.request_early_schedule_prob:
            doctor_information, result_dict = self.rescheduling_request(
                doctor_information=doctor_information,
                environment=environment, 
                verbose=verbose
            )
            if result_dict is not None:
                agent_test_data['doctor'] = doctor_information
                results['gt'].extend(result_dict['gt'])
                results['pred'].extend(result_dict['pred'])
                results['status'].extend(result_dict['status'])
                results['status_code'].extend(result_dict['status_code'])
                results['dialog'].extend(result_dict['dialog'])

                if verbose:
                    log(f'Pred  : {result_dict["pred"]}')
                    log(f'Status: {result_dict["status_code"]}')
                    log(f'Final Status: {result_dict["status_code"]}\n\n\n')

        return results



import requests
from typing import Optional
from urllib.parse import urlencode

from h_adminsim.utils import log



class FHIRManager:
    def __init__(self, fhir_url):
        self.fhir_url = fhir_url
        
    
    def __logging(self, response: requests.Response, verbose=True) -> Optional[requests.Response]:
        """
        Log the response status code and content.

        Args:
            response (requests.Response): The HTTP response object.
            verbose (bool, optional): If True, log details. Defaults to True.

        Returns:
            Optional[requests.Response]: The response object if JSON parsing is successful, else None.
        """
        if  200 <= response.status_code < 300:
            if verbose:
                log(f'Status code: {response.status_code}', color=True)
        else:
            if verbose:
                log(f'Status code: {response.status_code}', level='error')
        
        try:
            response_json = response.json()
            if verbose:
                log(f'Response JSON: {response_json}')      
        except ValueError:
            if verbose:
                log(f'Response Text: {response.text}', level='error')
            response = None
        return response


    def create(self, 
               resource_type: str, 
               resource_data: dict, 
               headers: Optional[dict] = None, 
               verbose: bool = True) -> Optional[requests.Response]:
        """
        Create a FHIR resource of the specified type.

        Args:
            resource_type (str): FHIR resource type (e.g., "Patient", "PractitionerRole").
            resource_data (dict): FHIR resource data as a dictionary.
            headers (Optional[dict], optional): HTTP headers to use. Defaults to None.
            verbose (bool, optional): If True, log details. Defaults to True.

        Returns:
            Optional[requests.Response]: The HTTP response object if JSON parsing is successful, else None.
        """
        _id = resource_data.get('id')
        fhir_url = f'{self.fhir_url}/{resource_type}/{_id}'
        response = requests.put(
            fhir_url,
            headers={'Content-Type': 'application/fhir+json'} if headers is None else headers,
            json=resource_data,
        )

        # Log and return the response
        return self.__logging(response, verbose)
    
    
    def read(self, 
             resource_type: str, 
             id: str, 
             headers: Optional[dict] = None, 
             verbose: bool = True) -> Optional[requests.Response]:
        """
        Read a FHIR resource of the specified type and ID.

        Args:
            resource_type (str): FHIR resource type (e.g., "Patient", "PractitionerRole").
            id (str): The ID of the FHIR resource to read.
            headers (Optional[dict], optional): HTTP headers to use. Defaults to None.
            verbose (bool, optional): If True, log details. Defaults to True.

        Returns:
            Optional[requests.Response]: The HTTP response object if JSON parsing is successful, else None.
        """
        fhir_url = f'{self.fhir_url}/{resource_type}/{id}'
        response = requests.get(
            fhir_url,
            headers={'Accept': 'application/fhir+json'} if headers is None else headers,
        )

        # Log and return the response
        return self.__logging(response, verbose)
    

    def update(self, 
               resource_type: str, 
               id: str, 
               resource_data: dict, 
               headers: Optional[dict] = None,
               verbose: bool = True) -> Optional[requests.Response]:
        """
        Update a FHIR resource of the specified type and ID.

        Args:
            resource_type (str): FHIR resource type (e.g., "Patient", "PractitionerRole").
            id (str): The ID of the FHIR resource to update.
            resource_data (dict): FHIR resource data as a dictionary.
            headers (Optional[dict], optional): HTTP headers to use. Defaults to None.
            verbose (bool, optional): If True, log details. Defaults to True.

        Returns:
            Optional[requests.Response]: _description_
        """
        fhir_url = f'{self.fhir_url}/{resource_type}/{id}'
        response = requests.put(
            fhir_url,
            headers={'Content-Type': 'application/fhir+json'} if headers is None else headers,
            json=resource_data,
        )

        # Log and return the response
        return self.__logging(response, verbose)


    def delete(self, resource_type: str, id: str, verbose=True):
        fhir_url = f'{self.fhir_url}/{resource_type}/{id}'
        response = requests.delete(
            fhir_url
        )

        # Log and return the response
        return self.__logging(response, verbose)


    def read_all(self,
                 resource_type: str,
                 headers: Optional[dict] = None,
                 count: int = 100,
                 verbose: bool = True,
                 params: Optional[dict] = None) -> list[dict]:
        """
        Read all resources of a given resource type using FHIR search with optional filtering.

        Args:
            resource_type (str): FHIR resource type (e.g., "PractitionerRole").
            headers (dict, optional): HTTP headers to use.
            count (int): Number of resources to fetch per page (default: 100).
            verbose (bool): If True, log each response. Defaults to True.
            params (dict, optional): FHIR search parameters (e.g., {"specialty": "IMALL-2"}).

        Returns:
            list[dict]: List of bundle entry dicts.
        """
        all_entries = []
        headers = {'Accept': 'application/fhir+json'} if headers is None else headers

        # Build first page URL with params
        q = {'_count': count}
        if params:
            q.update({k: v for k, v in params.items() if v is not None})
        url = f"{self.fhir_url}/{resource_type}?{urlencode(q, doseq=True)}"

        while url:
            response = requests.get(url, headers=headers)
            self.__logging(response, verbose)
            try:
                bundle = response.json()
            except Exception:
                break

            if bundle.get('resourceType') != 'Bundle' or 'entry' not in bundle:
                break

            all_entries.extend(bundle['entry'])

            # Check for next link (pagination)
            next_link = next(
                (link.get('url') for link in bundle.get('link', []) if link.get('relation') == 'next'),
                None
            )
            url = next_link  # Continue if next page exists, else break

        return all_entries
    

    def delete_all(self, entry: list[dict], verbose: bool = True):
        """
        Delete all FHIR resources from a given list of resource entries.

        Args:
            entry (list[dict]): List of FHIR Bundle entries, typically from the `read_all()` method.
                                Each entry should contain a 'resource' dict with 'resourceType' and 'id'.
            verbose (bool): If True, log each deletion response. Defaults to True.
        """
        error_ids = list()

        for resource in entry:
            resource_type = resource.get('resource').get('resourceType')
            id = resource.get('resource').get('id')
            response = self.delete(resource_type, id, verbose)

            if not 200 <= response.status_code < 300:
                error_ids.append(id)

        if error_ids:
            log(f'Error(s) occurs during delete resources: {error_ids}', 'warning')
        else:
            log('Deletion successfully completed', color=True)


        

# PostgreSQL
# docker exec -it jmlee_fhir_db psql -U admin -d hapi
# SELECT * FROM hfj_resource WHERE res_type = 'Patient' LIMIT 1;
# SELECT * FROM HFJ_RES_VER WHERE RES_ID = 925754;

    


import os
import random
from tqdm import tqdm
from importlib import resources
from typing import Optional, Tuple
from decimal import Decimal, getcontext

from h_adminsim.task.schedule_assign import ScheduleAssigner
from h_adminsim.utils import Information, log, colorstr
from h_adminsim.utils.common_utils import *
from h_adminsim.utils.filesys_utils import json_load, txt_load, yaml_save, make_project_dir, json_save_fast
from h_adminsim.utils.random_utils import (
    generate_random_prob,
    generate_random_date,
    generate_random_code,
    generate_random_names,
    generate_random_address,
    generate_random_telecom,
    generate_random_id_number,
    generate_random_specialty,
    generate_random_code_with_prob,
)



class DataSynthesizer:
    def __init__(self, config):
        # Initialize configuration, path and save directory
        self.config = config
        self._n = self.config.hospital_data.hospital_n
        self._save_dir = make_project_dir(self.config)
        self._data_save_dir = self._save_dir / 'data'
        yaml_save(self._save_dir / 'args.yaml', self.config)
        os.makedirs(self._data_save_dir, exist_ok=True)
        getcontext().prec = 10
        
    
    def synthesize(self,
                   return_obj: bool = False,
                   sanity_check: bool = False) -> Tuple[list[Information], list[Hospital]]:
        """
        Synthesize hospital data based on the configuration settings.

        Args:
            return_obj (bool, optional): Whether to return the hospital data object.
            sanity_check (bool, optional): If you want to check whether the generated data are compatible with the `Hospital` object,
                                 you can use this option.

        Raises:
            e: Exception if data synthesis fails.

        Returns:
            Tuple[list[Information], list[Hospital]]: A tuple containing the synthesized hospital data as an Information object and a Hospital object.
        """
        if sanity_check:
            return_obj = True

        try:
            all_data, all_hospitals = list(), list()
            hospitals = DataSynthesizer.hospital_list_generator(self.config.hospital_data.hospital_n)
            for i, hospital in tqdm(enumerate(hospitals), desc='Synthesizing data..', total=len(hospitals)):
                data = DataSynthesizer.define_hospital_info(self.config, hospital)
                hospital_obj = convert_info_to_obj(data) if return_obj else None
                if sanity_check:
                    new_data = convert_obj_to_info(hospital_obj)
                    assert to_dict(data) == to_dict(new_data)
                json_save_fast(self._data_save_dir / f'hospital_{padded_int(i, len(str(self._n)))}.json', to_dict(data))
                all_data.append(data)
                all_hospitals.append(hospital_obj)
            log(f"Total {len(hospitals)} data synthesizing completed. Path: `{self._data_save_dir}`", color=True)
            return all_data, all_hospitals
        
        except Exception as e:
            log(f"Data synthesizing failed: {e}", level='error')
            raise e


    @staticmethod
    def define_hospital_info(config, hospital_name: str) -> Information:
        """
        Define the synthetic hospital data, including its departments and doctors.

        Args:
            config: Configuration object containing hospital data settings.
            hospital_name (str): Name of the hospital to be defined.

        Returns:
            Information: Synthetic data about the hospital.
        """
        # Define hosptial metadata
        days = config.hospital_data.days
        dates = generate_date_range(
            generate_random_iso_date_between(
                str(config.hospital_data.start_date.min),
                str(config.hospital_data.start_date.max),
            ), 
            days
        )
        interval_hour = float(config.hospital_data.interval_hour)
        start_hour = float(random.randint(config.hospital_data.start_hour.min, config.hospital_data.start_hour.max))
        end_hour = float(random.randint(config.hospital_data.end_hour.min, config.hospital_data.end_hour.max))
        operation_hour_per_day = int(end_hour - start_hour)
        department_n = random.randint(
            config.hospital_data.department_per_hospital.min,
            config.hospital_data.department_per_hospital.max
        )
        doctor_n_per_department = [random.randint(config.hospital_data.doctor_per_department.min, config.hospital_data.doctor_per_department.max) 
                                   for _ in range(department_n)]
        doctor_n = sum(doctor_n_per_department)
        doctor_capacity_per_hour_list = [c for c in range(config.hospital_data.doctor_capacity_per_hour.min, config.hospital_data.doctor_capacity_per_hour.max + 1) \
                                         if float(Decimal(str(1))/Decimal(str(c)) % Decimal(str(interval_hour))) == 0]
        hospital_time_segments = convert_time_to_segment(start_hour, end_hour, interval_hour)
        metadata = Information(
            hospital_name=hospital_name,
            start_date=dates[0],
            end_date=dates[-1],
            days=days,
            department_num=department_n,
            doctor_num=doctor_n,
            time=Information(
                start_hour=start_hour,
                end_hour=end_hour,
                interval_hour=interval_hour
            )
        )

        # Define ScheduleAssigner class to randomly assign schedules to each doctor
        scheduler = ScheduleAssigner(start_hour, end_hour, interval_hour)

        # Define detailed hospital department, doctoral, and patient information
        department_info, doctor_info, patient_info = dict(), dict(), dict()
        departments = DataSynthesizer.department_list_generator(department_n)
        doctors = DataSynthesizer.name_list_generator(doctor_n, prefix='Dr. ')   # Doctor names are unique across all departments
        for department_data, doc_n in zip(departments, doctor_n_per_department):
            department, dep_code = department_data

            # Add department information
            department_info[department] = {'code': dep_code if dep_code else 'NA', 'doctor': []}
            
            # Add doctor information
            for _ in range(doc_n):
                doctor = doctors.pop()
                department_info[department]['doctor'].append(doctor)
                specialty, spe_code = generate_random_specialty(department)
                capacity_per_hour = random.choice(doctor_capacity_per_hour_list)
                working_days = random.randint(
                    config.hospital_data.working_days.min,
                    config.hospital_data.working_days.max
                )
                working_dates = sorted(random.sample(dates, working_days))
                doctor_info[doctor] = {
                    'department': department,
                    'specialty': {
                        'name': specialty,
                        'code': spe_code,
                    },
                    'schedule': {},
                    'capacity_per_hour': int(capacity_per_hour),
                    'capacity': int(capacity_per_hour * operation_hour_per_day * len(working_dates)),
                    'gender': generate_random_code('gender'),
                    'telecom': [{
                        'system': 'phone',
                        'value': generate_random_telecom(),
                        'use': generate_random_code('use')
                    }],
                    'birthDate': generate_random_date()
                }
                duration = int(1 / capacity_per_hour / interval_hour)

                # Generate doctor schedules and apponitments based on the pre-defined days
                for date in dates:
                    # Working day case
                    if date in working_dates:
                        schedule_segments, schedule_times = scheduler(
                            generate_random_prob(
                                config.hospital_data.doctor_has_schedule_prob,
                                config.hospital_data.schedule_coverage_ratio.min,
                                config.hospital_data.schedule_coverage_ratio.max
                            )
                        )
                        doctor_info[doctor]['schedule'][date] = schedule_times
                    # Not working day case
                    else:
                        schedule_segments, schedule_times = scheduler(1)
                        doctor_info[doctor]['schedule'][date] = schedule_times

                    # Add patient information per doctor
                    patient_segments = list(set(hospital_time_segments) - set(sum(schedule_segments, [])))
                    _, appointments = scheduler(
                        generate_random_prob(
                            1,
                            config.hospital_data.appointment_coverage_ratio.min,
                            config.hospital_data.appointment_coverage_ratio.max
                        ),
                        True,
                        patient_segments,
                        min_chunk_size=duration,
                        max_chunk_size=duration
                    )
                    patients = DataSynthesizer.name_list_generator(len(appointments))
                    for patient, appointment in zip(patients, appointments):
                        preference = generate_random_code_with_prob(
                            config.hospital_data.preference.type,
                            config.hospital_data.preference.probs
                        )
                        preference_rank = DataSynthesizer.second_preference_generator(preference)
                        symptom_level = generate_random_code_with_prob(
                            config.hospital_data.symptom.type,
                            config.hospital_data.symptom.probs
                        )
                        birth_date =  generate_random_date()
                        patient_info[patient] = {
                            'department': department,
                            'attending_physician': doctor,
                            'date': date,
                            'schedule': appointment,
                            'preference': preference_rank,
                            'symptom_level': symptom_level,
                            'gender': generate_random_code('gender'),
                            'telecom': [{
                                'system': 'phone',
                                'value': generate_random_telecom(),
                                'use': generate_random_code('use')
                            }],
                            'birthDate': birth_date,
                            'identifier': [{
                                'value': generate_random_id_number(birth_date=birth_date),
                                'use': 'official'

                            }],
                            'address': [{
                                'type': 'postal',
                                'text': generate_random_address(),
                                'use': 'home'
                            }]
                        }
            
        # Finalize data structure
        data = Information(
            metadata=metadata,
            department=department_info,
            doctor=doctor_info,
            patient=patient_info,
        )

        # Data sanity check
        if len(data.department) != metadata.department_num:
            raise AssertionError(colorstr('red', 'Department number mismatch'))
        if len(data.department) != len(set(doc['department'] for doc in data.doctor.values())):
            raise AssertionError(colorstr('red', 'Department number mismatch'))
        if len(data.doctor) != metadata.doctor_num:
            raise AssertionError(colorstr('red', 'Doctor number mismatch'))
        if len(data.doctor) != sum(len(dept['doctor']) for dept in data.department.values()):
            raise AssertionError(colorstr('red', 'Doctor number mismatch'))
        
        return data


    @staticmethod
    def hospital_list_generator(hospital_n: int,
                                file_path: Optional[str] = None) -> list[str]:
        """
        Generate a list of hospital names based on the number of hospitals.
        
        Args:
            hospital_n (int): Number of hospitals to generate.
            file_path (Optional[str], optional): Path to a file containing hospital names. If provided, it will be used to load names.
        
        Returns:
            list[str]: List of hospital names in the format "Hospital 001", "Hospital 002", etc.
        """
        if file_path:
            if registry.HOSPITALS is None:
                registry.HOSPITALS = [word.capitalize() for word in txt_load(file_path).split('\n') if word.strip()]
            return [f"{random.choice(registry.HOSPITALS)}" for _ in range(hospital_n)]
        
        zfill_l = len(str(hospital_n))
        return [f"hospital_{padded_int(i, zfill_l)}" for i in range(hospital_n)]

    
    @staticmethod
    def department_list_generator(department_n: int,
                                  file_path: Optional[str] = None) -> list[Tuple[str, str]]:
        """
        Generate a list of department names based on the number of departments.
        
        Args:
            department_n (int): Number of departments to generate.
            file_path (Optional[str], optional): Path to a file containing department names. If provided, it will be used to load names. Defaults to None.
        
        Returns:
            list[Tuple[str, str]]: List of department names and their codes.
        """
        if file_path == None:
            file_path = str(resources.files("h_adminsim.assets.departments").joinpath("department.json"))

        if file_path:
            if registry.DEPARTMENTS is None:
                specialty = json_load(file_path)['specialty']
                registry.DEPARTMENTS = [(k2, v2['code']) for v1 in specialty.values() for k2, v2 in v1['subspecialty'].items()]
            
            if department_n > len(registry.DEPARTMENTS):
                raise ValueError(f"Requested {department_n} departments, but only {len(registry.DEPARTMENTS)} available in {file_path}.")
        
            return random.sample(registry.DEPARTMENTS, department_n)
            
        zfill_l = len(str(department_n))
        return [(f"department_{padded_int(i, zfill_l)}", None) for i in range(department_n)]
    
    
    @staticmethod
    def name_list_generator(n: int,
                            first_name_file_path: Optional[str] = None, 
                            last_name_file_path: Optional[str] = None,
                            prefix: Optional[str] = None) -> list[str]:
        """
        Generate a list of names.
        
        Args:
            n (int): Number of doctors to generate.
            first_name_file_path (Optional[str], optional): Path to a file containing first names. Defaults to None.
            last_name_file_path (Optional[str], optional): Path to a file containing last names. Defaults to None.
            prefix (Optional[str], optional): Prefix for to be generated names.
        
        Returns:
            list[str]: List of names.
        """
        if first_name_file_path == None:
            first_name_file_path = str(resources.files("h_adminsim.assets.names").joinpath("firstname.txt"))
        if last_name_file_path == None:
            last_name_file_path = str(resources.files("h_adminsim.assets.names").joinpath("lastname.txt"))

        if prefix != None:
            assert isinstance(prefix, str), log("`prefix` must be a string type", "error")
            names = [f'{prefix}{name}' for name in generate_random_names(n, first_name_file_path, last_name_file_path)]
        else:
            names = [name for name in generate_random_names(n, first_name_file_path, last_name_file_path)]
        random.shuffle(names)
        return names


    @staticmethod
    def second_preference_generator(preference: str) -> list[str]:
        """
        Generate a list of preferences based on the initial preference.

        Args:
            preference (str): First priority of preference.

        Returns:
            list[str]: List of preferences including first and second priority.
        """
        preference_list = [preference]
        
        if preference == 'doctor':
            second_preference = random.choice(['asap', 'date'])
            preference_list.append(second_preference)
        elif preference == 'date':
            second_preference = random.choice(['asap', 'doctor'])
            preference_list.append(second_preference)
        elif preference == 'asap':
            second_preference = random.choice(['date', 'doctor'])
            preference_list.append(second_preference)

        return preference_list



import os
from tqdm import tqdm
from typing import Optional
from decimal import getcontext
from importlib import resources

from h_adminsim.utils.fhir_utils import *
from h_adminsim.utils.random_utils import generate_random_symptom
from h_adminsim.utils.filesys_utils import json_load, json_save_fast, get_files



class AgentDataBuilder:
    def __init__(self, config):
        # Initialize configuration
        data_dir = os.path.join(config.project, config.data_name, 'data')
        self.data_files = get_files(data_dir, ext='json')
        getcontext().prec = 10


    @staticmethod
    def build(data: dict, 
              save_path: Optional[str] = None, 
              symptom_file_path: Optional[str] = None) -> dict:
        """
        Build agent test data from a single hospital data entry.

        Args:
            data (dict): Dictionary containing metadata, departments, doctors, and patients.
            save_path (Optional[str], optional): If provided, the generated agent data will be saved to this path.
            symptom_file_path (str, optional): Path to the JSON file containing symptoms per department. Defaults to None.

        Returns:
            dict: A dictionary with the following keys:
                - 'metadata': Original metadata from input.
                - 'department': Department-level information.
                - 'doctor': Doctor-level information.
                - 'agent_data': List of tuples. Each tuple consists of:
                    - Ground-truth scheduling information.
                    - Agent input (symptom and constraints).
        """
        if symptom_file_path == None:
            symptom_file_path = str(resources.files("h_adminsim.assets.departments").joinpath("symptom.json"))

        agent_data = {'metadata': data['metadata'], 'department': data['department'], 'doctor': data['doctor'], 'agent_data': []}
        
        for patient, patient_values in data['patient'].items():
            doctor, department, date = patient_values['attending_physician'], patient_values['department'], patient_values['date']
            gender, telecom, birth_date, identifier, address = \
                patient_values['gender'], patient_values['telecom'], patient_values['birthDate'], patient_values['identifier'], patient_values['address']
            preference, symptom_level = patient_values['preference'], patient_values['symptom_level']
            disease = generate_random_symptom(
                department=department,
                symptom_file_path=symptom_file_path, 
                ensure_unique_department='doctor' in patient_values['preference']
            )
            gt_department = disease['department'] if isinstance(disease, dict) else [department]
            gt = {
                'patient': patient,
                'gender': gender,
                'telecom': telecom,
                'birthDate': birth_date,
                'identifier': identifier,
                'address': address,
                'department': gt_department,
                'attending_physician': doctor,
                'valid_from': date if 'date' in preference else 'N/A',
                'preference': preference,
                'symptom_level': symptom_level,
            }
            agent = {
                'patient': patient,
                'gender': gender,
                'telecom': telecom,
                'birthDate': birth_date,
                'identifier': identifier,
                'address': address,
                'constraint': {
                    'preference': preference,
                    'attending_physician': doctor,
                    'valid_from': date if 'date' in preference else 'N/A',
                    'symptom_level': symptom_level,
                    'symptom': disease,
                }
            }
            agent_data['agent_data'].append((gt, agent))
        
        if save_path:
            json_save_fast(
                save_path,
                agent_data
            )
        
        return agent_data
            

    def __call__(self,
                 output_dir: Optional[str] = None, 
                 symptom_file_path: Optional[str] = None) -> list[dict]:
        """
        Generate agent test datasets for all input data files.

        Args:
            output_dir (Optional[str], optional): Directory to save the generated agent data files.
                                                  If not provided, files are not saved.
            symptom_file_path (Optional[str], optional): Path to the symptom file used during agent construction. Defaults to None.

        Returns:
            list[dict]: A list of agent test data dictionaries, one for each processed input file.
        """
        if symptom_file_path == None:
            symptom_file_path = str(resources.files("h_adminsim.assets.departments").joinpath("symptom.json"))

        os.makedirs(output_dir, exist_ok=True)
        all_agent_data = list()
        
        for data_file in tqdm(self.data_files, desc='Generating data for agent simulation..'):
            data = json_load(data_file)
            basename, ext = os.path.splitext(os.path.basename(data_file))
            agent_data = AgentDataBuilder.build(data, os.path.join(output_dir, f"{basename}_agent{ext}"), symptom_file_path)
            all_agent_data.append(agent_data)
        
        return all_agent_data
        


from typing import Optional
from pydantic import BaseModel, Field



class PromptRequest(BaseModel):
    user_prompt: str
    system_prompt: Optional[str] = None



class ScheduleItem(BaseModel):
    start: float = Field(description="Start time of the appointment")
    end: float = Field(description="End time of the appointment")



class ScheduleModel(BaseModel):
    schedule: dict[str, ScheduleItem] = Field(description="Doctor's schedule")
    changed_existing_schedule_list: list = Field(description="List of changed appointments among the exised schedules")
    